# -*- coding: utf-8 -*-

"""

  model
  ~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

__version__ = 'v5'

# stdlib
import abc
import string
import inspect
import operator
import datetime

# model components
from . import query
from . import exceptions

# dateutil
from dateutil import parser as dtparser

# model adapters
from .adapter import abstract, concrete
from .adapter import KeyMixin, ModelMixin
from .adapter import VertexMixin, EdgeMixin

# datastructures
from canteen.util.struct import EMPTY
from canteen.util.struct import BidirectionalEnum


# Globals / Sentinels
_NDB = False  # `canteen.model` no longer supports NDB
_MULTITENANCY = False  # toggle multi-tenant key namespaces
Edge = Vertex = Model = AbstractModel = None  # must initially be `None`
_DEFAULT_KEY_SCHEMA = ('id', 'kind', 'parent')  # default key schema
_MULTITENANT_KEY_SCHEMA = ('id', 'kind', 'parent', 'namespace', 'app')
_BASE_MODEL_CLS = frozenset(('Model', 'AbstractModel', 'Vertex', 'Edge'))
_BASE_GRAPH_CLS = frozenset(('Vertex', 'Edge'))
is_adapter = lambda _t: issubclass(_t.__class__, abstract.ModelAdapter)
_PROPERTY_SLOTS = (
  'name',  # property string name (used as attribute/item key)
  '_options',  # arbitrary userland property options
  '_indexed',  # should this property be indexed?
  '_required',  # should property should be required?
  '_repeated',  # should property be allowed to keep multiple values?
  '_basetype',  # property base type, for validation and storage
  '_default')  # default value for property, defaults to ``None``


def is_there_a_global(name):

  """ Simple utility to interrogate the global context
      and see if something is defined yet.

      :param name: Name to check for global definition
        in this module.

      :returns: Whether the target ``Name`` is defined
        in module globals and is not falsy. """

  gl = globals()
  return name in gl and (not gl[name])


# == Metaclasses == #

class MetaFactory(type):

  """ Abstract parent for Model API primitive metaclasses,
      such as :py:class:`AbstractKey.__metaclass__` and
      :py:class:`AbstractModel.__metaclass__`.

      Enforces the metaclass chain and proper :py:mod:`abc`
      compliance.

      .. note :: Metaclass implementors of this class **must**
             implement :py:meth:`cls.initialize()`, or
             :py:class:`Model` construction will yield a
             :py:exc:`NotImplementedError`. """

  class __metaclass__(abc.ABCMeta):

    """ Embedded metaclass - enforces ABC compliance and
      properly formats :py:attr:`cls.__name__`. """

    __owner__ = 'MetaFactory'

    def __new__(mcs, name=None, bases=tuple(), properties=None):

      """ Factory for metaclasses classes. Regular
          metaclass factory function.

          If the target class definition has the attribute
          :py:attr:`cls.__owner__`, it will be taken as the
          target class' internal ``__name__``. ``basestring``
          and classes are accepted (in which case the bound
          class' name is taken instead).

          :param name: String name for the metaclass class to factory.
          :param bases: Metaclass class inheritance path.
          :param properties: Property dictionary, as defined inline.
          :returns: Constructed :py:class:`MetaFactory.__metaclass__`
            descendant.

          .. note:: This class is *two* levels up in the meta chain.
                Please note this is an *embedded* metaclass used for
                *metaclass classes*. """

      # alias embedded metaclasses to their `__owner__` (for __repr__)
      name = mcs.__name__ = mcs.__owner__ if hasattr(mcs, '__owner__') else name

      # enforces metaclass
      return super(mcs, mcs).__new__(mcs, name, bases, properties or {})

  # = Internal Methods = #
  def __new__(mcs, name=None, bases=tuple(), properties=None):

    """ Factory for concrete metaclasses. Enforces
        abstract-ness (prevents direct construction) and
        dispatches :py:meth:`mcs.initialize()`.

        :param name: String name for the metaclass to factory.
        :param bases: Inheritance path for the new concrete metaclass.
        :param properties: Property dictionary, as defined inline.
        :returns: Constructed :py:class:`MetaFactory` descendant.
        :raises: :py:exc:`model.exceptions.AbstractConstructionFailure`
          upon concrete construction. """

    # fail on construction - embedded metaclasses cannot be instantiated
    if not name: raise exceptions.AbstractConstructionFailure(mcs.__name__)

    # pass up to `type`, which properly enforces metaclasses
    impl = mcs.initialize(name, bases, properties or {})

    # if we're passed a tuple, we're being asked to super-instantiate
    if isinstance(impl, tuple):
      return super(MetaFactory, mcs).__new__(mcs, *impl)
    return impl

  ## = Exported Methods = ##
  @classmethod
  def resolve(mcs, name, bases, properties, default=True):

    """ Resolve a suitable model adapter for a given model Key or Model.

        :param name: Class name, as provided to :py:meth:`__new__`.
        :param bases: Inheritance path for the target :py:class:`Model`.
        :param properties: Class definition, as provided to :py:meth:`__new__`.
        :keyword default: Allow use of the default adapter. Defaults to `True`.
        :returns: A suitable :py:class:`model.adapter.ModelAdapter` subclass.

        :raises NoSupportedAdapters: in the case that no supported (or valid)
          adapters could be found.

        :raises InvalidExplicitAdapter: in the case of an unavailable,
          explicitly-requested adapter. """

    from canteen.model.adapter import abstract

    if '__adapter__' not in properties:

      for i in bases:
        if hasattr(i, '__adapter__'):
          # parent specifies adapter and no explicit, so use that
          return None, i.__adapter__.acquire(name, bases, properties)

      # grab each supported adapter
      _adapters = [option for option in concrete if option.is_supported()]

      # fail with no adapters...
      if not _adapters: raise exceptions.NoSupportedAdapters()

      # if we only have one adapter, the choice is easy...
      if not default:
        return _adapters[0].acquire(name, bases, properties), tuple()
      return _adapters[0].acquire(name, bases, properties)

    # an explicit adapter was requested via an `__adapter__` class property
    _spec = properties['__adapter__']

    if not isinstance(_spec, (list, tuple)):
      _spec = [_spec]

    for _spec_item in _spec:

      if isinstance(_spec_item, (type, basestring)):
        for _a in concrete:
          if _a is _spec_item or _a.__name__ == _spec_item:
            return _a.acquire(name, bases, properties)

      elif is_adapter(_spec_item):  # pragma: no cover
        # it's a valid model adapter instance already
        return _spec_item  # pragma: no cover

    adapter_label = _spec
    if not isinstance(_spec, basestring):  # pragma: no cover
      if isinstance(_spec, type):
        adapter_label = _spec.__name__
      if issubclass(_spec.__class__, abstract.ModelAdapter):
        adapter_label = _spec.__class__.__name__
    raise exceptions.InvalidExplicitAdapter(adapter_label)

  ## = Abstract Methods = ##
  @abc.abstractmethod
  def initialize(cls, name, bases, properties):

    """ Initialize a subclass. Must be overridden by child metaclasses.

        :param name: Target class name to initialize.
        :param bases: Target class bases to mix in.
        :param properties: Bound properties to target class.

        :raises NotImplementedError: Always, if called directly, as this
          method is abstract. """

    raise NotImplementedError('`initialize` is abstract.')  # pragma: no cover


## == Abstract Classes == ##

class AbstractKey(object):

  """ Abstract Key class. Provides base abstract
      functionality supporting concrete Key objects. """

  __schema__ = _DEFAULT_KEY_SCHEMA  # set default key schema

  ## = Encapsulated Classes = ##
  class __metaclass__(MetaFactory):

    """ Metaclass for model keys. Reorganizes internal
        class structure to support object-specific
        behavior. See ``cls.initialize`` for details. """

    # class owner and schema to spawn by default
    __owner__, __schema__ = 'Key', _DEFAULT_KEY_SCHEMA

    @classmethod
    def initialize(mcs, name, bases, pmap):

      """ Initialize a Key class. Reorganize class layout
          to support ``Key`` operations. The class property
          ``__schema__`` (expected to be an iterable of string
          schema item names) is scanned for key structure.

          The following non-standard class-level properties are set:
            - ``__slots__`` - frozen to nothing as everything is class-level
            - ``__owner__`` - owner model object for a key, if any
            - ``__adapter__`` - adapter for this key, resolved at class time
            - ``__persisted__`` - whether this key is known to be persisted

          Each schema'd key chunk is also given an internal-style
          property (read: surrounded by ``__``).

          :param name: Name of ``Key`` class to initialize.
          :param bases: Base classes to mix into target ``Key`` class.
          :param pmap: Map of properties bound to target ``Key`` class.

          :returns: Constructed type extending ``Key``. """

      # resolve adapter
      _adapter = mcs.resolve(name, bases, pmap)
      _module = pmap.get('__module__', 'canteen.model')

      if isinstance(_adapter, tuple):
        _base, _adapter = _adapter
      else:
        _base, _adapter = False, _adapter  # use concrete adapter

      if name == 'AbstractKey':  # <-- must be a string
        return name, bases, dict([('__adapter__', _adapter)] + pmap.items())

      key_class = [  # build initial key class structure
        ('__slots__', set()),  # seal object attributes
        ('__bases__', bases),  # define bases for class
        ('__name__', name),  # set class name internals
        ('__owner__', None),  # reference to current owner entity
        ('__module__', _module),  # add class package
        ('__persisted__', False)]  # default to not persisted

      if _base is not None:  # allow adapter definition to defer upwards
        key_class.append(('__adapter__', _adapter))  # pragma: no cover

      # resolve schema and add key format items, initted to None
      _schema = [
        ('__%s__' % x, None) for x in pmap.get('__schema__', mcs.__schema__)]

      # return an argset for `type`
      return name, bases, dict(_schema + key_class + pmap.items())

    def mro(cls):

      """ Generate a fully-mixed MRO for `AbstractKey` subclasses.
          Handles injection of ``CompoundKey`` object in MRO to support
          ``KeyMixin`` composure.

          The non-base MRO pattern is as follows:
            - for ``AbstractKey``: ``KeyMixin -> object``
            - for ``Key``: `AbstractKey -> KeyMixin -> object``
            - for ``Key`` classes: ``Key -> AbstractKey -> KeyMixin -> object``

          :returns: Properly-mixed MRO for an `AbstractKey` subclass. """

      if cls.__name__ == 'AbstractKey':  # `AbstractKey` MRO
        return (cls, KeyMixin.compound, object)

      if cls.__name__ == 'Key':  # `Key` MRO
        return (cls, AbstractKey, KeyMixin.compound, object)

      # `Key`-subclass MRO, with support for diamond inheritance
      return tuple(filter(lambda x: x not in (Key, AbstractKey), [cls] + (
        list(cls.__bases__))) + [Key, AbstractKey, KeyMixin.compound, object])

    def __repr__(cls):

      """ String representation of a `Key` class. Note that this
          ``__repr__`` handles ``cls`` representation.

          :returns: ``x(y)`` where ``x`` is the `Key` class in use
            and ``y`` is collapsed key schema. """

      # dump key schema
      return '%s(%s)' % (
        cls.__name__, ', '.join((i for i in reversed(cls.__schema__))))

  def __new__(cls, *args, **kwargs):

    """ Intercepts construction requests for abstract model classes.
        Enforces abstractness, but otherwise passes calls to ``super``.

        :raises AbstractConstructionFailure: In the case that an attempt
          is made to construct an ``AbstractKey`` directly.

        :returns: New instance of `cls`, so long as `cls` it is concrete. """

    # prevent direct instantiation of `AbstractKey`
    if cls.__name__ == 'AbstractKey':
      raise exceptions.AbstractConstructionFailure('AbstractKey')

    return super(cls, cls).__new__(*args, **kwargs)  # pragma: no cover

  def __eq__(self, other):

    """ Test whether two keys are functionally identical.

        Tries the following tests to negate equality:
        - target vs self falsyness
        - target vs self schema
        - target vs self type
        - target vs self data

        If all elements are reported to be equal, the keys
        are functionally identical. If any of the tests
        fail, the keys are not identical.

        :param other: Other key to test against ``self``.

        :returns: ``True`` if ``other`` is functionally
          identical to ``self``, otherwise ``False``. """

    if (not self and not other) or (self and other):
      if isinstance(other, self.__class__):  # class check
        if self.__schema__ == other.__schema__:
            # last resort: check each data property
            return all((i for i in map(lambda x: (
                getattr(other, x) == getattr(self, x)), self.__schema__)))
    # didn't pass one of our tests
    return False  # pragma: no cover

  def __repr__(self):

    """ Generate a string representation of this Key object.
        Note that this method is responsible for ``self``
        representation, not ``cls``.

        :returns: String describing the local ``Key`` object,
          like ``x(y)``, where ``x`` is the local ``Key``
          class name and ``y`` is a formatted list of this
          keys's schema and data. """

    pairs = ('%s=%s' % (k, getattr(self, k)) for k in reversed(self.__schema__))
    return "%s(%s)" % (self.__class__.__name__, ', '.join(pairs))

  # util: alias `__repr__` to string magic methods
  __str__ = __unicode__ = __repr__

  # util: support for `__nonzero__` and aliased `__len__`
  __nonzero__ = lambda self: isinstance(self.__id__, (basestring, int, long))
  __len__ = lambda self: (
    int(self.__nonzero__()) if self.__parent__ is None else (
      sum((1 for i in self.ancestry))))

  ## = Property Setters = ##
  def _set_internal(self, name, value):

    """ Set an internal property on a `Key`. Used by ``Key``
        internals to proxy attribute and item calls.

        :param name: Name of the internal property to set.
        :param value: Value to set the internal property to.

        :raises PersistedKey: If an internal property set is
          attempted on an already-persisted ``Key``.

        :returns: ``self``, for chainability. """

    if value is '': value = None

    # fail if we're already persisted (unless we're updating the owner)
    if self.__persisted__ and name != 'owner':
      raise exceptions.PersistedKey(name)
    if name == 'parent' and isinstance(value, Model):
      # special case: if setting parent and it's a model, use the key instead
      value = value.key  # pragma: no cover
    if name == 'id' and isinstance(value, basestring) and ':' in value:
      raise ValueError('Keynames may not contain the ":" character.'
                       ' Got: "%s".' % value)  # pragma: no cover
    setattr(self, '__%s__' % name, value)
    return self

  ## = Property Getters = ##
  def _get_ancestry(self):

    """ Retrieve this ``Key``'s ancestry path, one item at a
        time, starting from this ``Key``'s root and proceeding
        down the chain. Acts as a generator.

        :raises StopIteration: When no more ancestry entries
          are available to yield.

        :returns: Yields ancestry items one at a time, starting
          with this ``Key``'s root, and ending with ``self``. """

    if self.__parent__:  # if we have a parent, yield upward
      for i in self.__parent__.ancestry: yield i
    yield self  # yield self to end the chain, and stop iteration
    raise StopIteration()

  ## = Property Bindings  = ##
  id = property(lambda self: self.__id__,
                lambda self, id: self._set_internal('id', id))

  app = property(lambda self: self.__app__,
                 lambda self, app: self._set_internal('app', app))

  kind = property(lambda self: self.__kind__,
                  lambda self, kind: self._set_internal('kind', kind))

  parent = property(lambda self: self.__parent__,
                    lambda self, p: self._set_internal('parent', p))

  owner = property(lambda self: self.__owner__)  # `owner` is read-only
  ancestry = property(_get_ancestry)  # `ancestry` is read-only

  # ns = `None` when disabled
  namespace = property(lambda self: self.__namespace__ or None,
                       lambda self, ns: self._set_internal('namespace', ns))


## AbstractModel
class AbstractModel(object):

  """ Abstract Model class. Specifies abstract
      properties that apply to all `Model`-style
      data containers.

      Applies an embedded `MetaFactory`-compliant
      metaclass to properly initialize subclasses. """

  __slots__ = tuple()


  ## = Encapsulated Classes = ##
  class __metaclass__(MetaFactory):

    """ Embedded `MetaFactory`-compliant metaclass.
        Handles initialization of new `AbstractModel`
        descendent classes.

        Reorganizes class layout and structure to
        satisfy future internal model operations.
        See ``cls.initialize`` for details. """

    __owner__ = 'Model'

    @staticmethod
    def _get_prop_filter(inverse=False):

      """ Builds a callable that can filter properties
          between internal and user-defined ones. Returns

          By default, the callable produced by this
          method will produce ``False`` if the property
          is internal and ``True`` otherwise.

          To flip this functionality, pass ``True`` to
          inverse, which simply converts the boolean
          response to ``True`` for internal properties,
          and ``False`` for user data.

          :param inverse: Flip the filter to only return
            truthy for internal properties. Defaults to
            ``False`` for internal-filter behavior.

          :returns: Closure that is appropriate for use
            with ``filter`` and returns ``True`` if the
            property ``bundle`` handed in is internal. """

      def _filter_prop(bundle):

        """ Decide whether a property is kept as a data
            value or a class internal. See wrapping
            function for full description.

            :param bundle: Property bundle to examine.

            :returns: ``True`` or ``False`` according
              to the property ``bundle``'s status as
              an internal property. """

        key, value = bundle  # extract, this is dispatched from ``{}.items``
        if key.startswith('_'): return inverse
        if isinstance(value, classmethod): return inverse
        if inspect.isfunction(value) or inspect.ismethod(value): return inverse
        return (not inverse)

      return _filter_prop

    @classmethod
    def initialize(mcs, name, bases, properties):

      """ Initialize a ``Model`` descendent for use as a
          data model class. Reorganizes class internals
          to store descriptors for defined user data
          properties.

          Also injects a few properties into the class
          map to satisfy ``Model`` layer functionality.

          :param name: Name of the ``Model``-descendent
            class to initialize.

          :param bases: Base classes to extend with ``cls``
            target. Must include ``Model``, ``Vertex``,
            ``Edge`` or a class that extends one thereof.

          :param properties: Class property map, as defined
            inline or via a manual call to ``initialize``.
            Always expected to be a ``dict``.

          :raises AssertionError: If incorrect types are
            passed for ``name``, ``bases`` or ``properties``.
            ``__debug__`` must be active to enable assertions.

          :returns: Initialized target ``Model``-descendent
            class, with rewritten property map. """

      assert isinstance(name, basestring), "class name must be a string"
      assert isinstance(bases, tuple), "class bases must be a tuple of types"
      assert isinstance(properties, dict), "class map must be a valid dict"

      property_map = {}
      _nondata_map = {}

      # core classes eval before being defined - must use string name :(
      if name not in frozenset(['AbstractModel', 'Model']):

        modelclass = {}

        # parse spec (`name=<basetype>` or `name=<basetype>,<options>`)
        # also, model properties that start with '_' are ignored
        for prop, spec in (
            filter(mcs._get_prop_filter(), properties.iteritems())):

          if any((char not in string.lowercase for char in prop[1:])) and (
              isinstance(spec, type) and (
                  issubclass(spec, BidirectionalEnum))):  # pragma: no cover
            modelclass[prop] = spec
            continue  # camel-case properties should be skipped for init

          # build a descriptor object and data slot
          basetype, options = (
            (spec, {}) if not isinstance(spec, tuple) else spec)

          # simple strings should accept both str and unicode
          if basetype is str: basetype = basestring

          # (note that specifying unicode explicitly only accepts unicode)
          if isinstance(basetype, Property):
            property_map[prop] = basetype  # pragma: no cover
          else:
            property_map[prop] = Property(prop, basetype, **options)

        # drop non-data-properties into our ``_nondata_map``
        for prop, value in filter(mcs._get_prop_filter(inverse=True),
                                  properties.iteritems()):
          _nondata_map[prop] = value

        # merge and clone all basemodel properties, update dictionary
        if len(bases) > 1 or bases[0] is not Model:

          # build a full property map, after reducing parents left -> right
          _pmap_data = ([(
            [(prop, b.__dict__[prop].clone()) for prop in b.__lookup__])
              for b in bases] + [property_map.items()])

          # calculate full dict property map
          property_map = dict([(key, value) for key, value in (
            reduce(operator.add, _pmap_data))])

        # freeze property lookup
        prop_lookup = frozenset((k for k, v in property_map.iteritems()))

        # resolve default adapter for model
        model_adapter = mcs.resolve(name, bases, properties)
        if isinstance(model_adapter, tuple):
          _base, model_adapter = model_adapter
        else:
          _base = False  # we have an adapter, use it

        _model_internals = {  # build class layout, initialize core model
          '__impl__': {},  # holds cached implementation classes
          '__name__': name,  # add internal class name (should be Model kind)
          '__kind__': name,  # kindname defaults to model class name
          '__bases__': bases,  # stores a model class's bases, so MRO can work
          '__lookup__': prop_lookup,  # frozenset of allocated attributes
          '__module__': properties.get('__module__'),  # add model's module
          '__slots__': tuple()}  # seal-off object attributes

        # resolves default adapter class, but allow deference upwards
        if _base is not None:
          _model_internals['__adapter__'] = model_adapter

        modelclass.update(property_map)  # update at class-level
        modelclass.update(_nondata_map)  # update at class-level with non data
        modelclass.update(_model_internals)  # lastly, apply model internals

        # for top-level graph objects
        if name in ('Edge', 'Vertex'):
          if is_there_a_global(name):
            # make flag for edge/vertex class
            graph_object_flag = '__%s__' % name.lower()

            # mark as graph model and according type
            properties['__graph__'] = properties[graph_object_flag] = True

        # for graph-object subclasses
        elif any((hasattr(b, '__graph__') for b in bases)) or (
                name in ('Edge', 'Vertex')):

          # find owner name
          object_owner = 'Vertex' if (
            any((hasattr(b, '__vertex__') for b in bases))) else 'Edge'

          # set flags
          graph_flag = '__%s__' % object_owner.lower()
          _model_internals['__owner__'] = object_owner
          _model_internals['__graph__'] = _model_internals[graph_flag] = True

        # inject our own property map
        impl = super(MetaFactory, mcs).__new__(mcs, name, bases, modelclass)
        return model_adapter._register(impl)

      return name, bases, properties  # pass-through to `type`

    def mro(cls):

      """ Generate a fully-mixed method resolution order for
          `AbstractModel` subclasses.

          According to the parent base that we're extending,
          models get custom MRO chains to fulfill mixin
          attributes in the same Canteen-style DI at runtime.

          :raises RuntimeError: If MRO cannot be calculated
            because the target ``cls`` is completely invalid.

          :returns: Mixed MRO according to base class structure. """

      def find_nonbase_mro():

        """ Finds proper MRO for non-base classes, i.e. classes
            that actually make use of ``Model``, ``Edge`` or
            ``Vertex`` (not those classes themselves).

            :returns: Class-tree-specific MRO, if possible, in a
              ``list`` suitable for composure into a final MRO. """

        base_models = frozenset((Vertex, Edge, Model, AbstractModel))

        # graph models?
        if any((hasattr(b, '__graph__') for b in cls.__bases__)):
          _vertex = any((hasattr(b, '__vertex__') for b in cls.__bases__))
          return [i for i in cls.__bases__ if i not in base_models] + [
                   Vertex if _vertex else Edge,
                   Model, AbstractModel,
                   VertexMixin.compound if _vertex else EdgeMixin.compound,
                   ModelMixin.compound]

        # vanilla non-base MRO
        return [i for i in cls.__bases__ if i not in base_models] + [
                  Model, AbstractModel, ModelMixin.compound]

      # calculate common `Model` MRO
      base_model_mro = (
        [ModelMixin.compound] if cls.__name__ == 'AbstractModel' else [
          AbstractModel, ModelMixin.compound])

      target_mro = {

        # `AbstractModel` just gets the mixin
        'AbstractModel': lambda: [ModelMixin.compound],

        # `Model` gets `AbstractModel` and mixin
        'Model': lambda: base_model_mro,

        # `Vertex` structure overrides `Model` structure
        'Vertex': lambda: [VertexMixin.compound, Model] + base_model_mro,

        # `Edge` structure overrides `Model` structure
        'Edge': lambda: [EdgeMixin.compound, Model] + base_model_mro,

      }.get(cls.__name__, find_nonbase_mro)()

      if not target_mro or not isinstance(target_mro, list):
        # not a base class
        raise RuntimeError('Unable to calculate MRO for unidentified'
                           ' MetaFactory-descendent'
                           ' Model class: %s.' % cls)  # pragma: no cover

      return tuple([cls] + target_mro + [object])

    def __repr__(cls):

      """ Generate string representation of `Model` class,
          like "Model(<prop1>, <prop n...>)".

          :returns: String representation of this `Model. """

      if hasattr(cls, '__lookup__') and cls.__name__ not in _BASE_MODEL_CLS:
        return '%s(%s)' % (cls.__name__, ', '.join((i for i in cls.__lookup__)))

      elif (cls.__name__ in _BASE_GRAPH_CLS and (
        cls.__owner__ in _BASE_GRAPH_CLS)) or (cls.__name__ == 'Model' or (
            cls.__name__ == 'AbstractModel')):
        return cls.__name__

      return 'Model<%s>' % cls.__name__  # pragma: no cover

    __str__ = __unicode__ = __repr__

    def __setattr__(cls, name, value,
                               exception=exceptions.InvalidAttributeWrite):

      """ Disallow property mutation before instantiation.

          The Exception raised for invalid writes can be set
          via ``exception``.

          :param name: Name of the property we are being asked to mutate.
          :param value: Value that we are being asked to set ``name`` to.
          :param exception: Exception to be raised in the event of an error.
            Handles switchover to ``KeyError``-based exceptions if this method
            is being called from Python's item-style API.

          :raises InvalidItem: In the case that this method is used via Python's
            Item API and an invalid attempt is made to write to a property.

          :raises InvalidAttributeWrite: In the case that this method is used
            in an invalid way to attempt an attribute write.

          :returns: Generally ``None`` if a successful write is made. """

      # cannot mutate data before instantiation
      if name in cls.__lookup__:
        raise exception('mutate', name, cls)

      # setting internal class-level properties
      if name.startswith('__') or callable(value) or (
        isinstance(value, (classmethod, staticmethod))):
        return super(AbstractModel.__metaclass__, cls).__setattr__(name, value)

      # cannot create new properties before instantiation
      raise exception('create', name, cls)

    def __getitem__(cls, name, exception=exceptions.InvalidItem):

      """ Override itemgetter syntax to return property
          objects at the class level.

          :param name: Name of the value we wish to retrieve.
          :param exception: Exception to be raised

          :raises InvalidItem: In the case of an attempt to read
            an invalid or non-existent property value.

          :returns: Value of the target property ``name``, or the
            :py:class:`model.Property` object itself if accessed
            at a class level. """

      if name not in cls.__lookup__:  # pragma: no cover
        raise exception('read', name, cls)  # cannot read non-data properties
      return cls.__dict__[name]


  class _PropertyValue(object):

    """ Named-tuple class for property value bundles.
        DOCSTRING"""

    __fields__, __slots__ = zip(*((f, '__%s__' % f) for f in ('data', 'dirty')))

    def __init__(self, data, dirty=False):

      """ Initialize a new `PropertyValue` instance.

          :param data: The data to store in this ``_PropertyValue`` instance.

          :param dirty: Whether the data has been modified/mutated since
            original instantiation. Defaults to ``False``, as the most common
            use case for constructing a new ``_PropertyValue`` instance is data
            coming back from an adapter. """

      self.__data__, self.__dirty__ = data, dirty

    # util: map data and dirty properties
    data, dirty = (property(lambda self: self.__data__),
                   property(lambda self: self.__dirty__))

    # util: reduce arguments for pickle
    __getnewargs__ = lambda self: tuple(self)

    # util: iterate over items in order
    __iter__ = lambda self: iter((self.__data__, self.__dirty__))

    # util: support getitem syntax, where 0 is `data` and 1 is `dirty`
    __getitem__ = lambda self, item: (
        self.__data__ if not item else self.__dirty__)

    # util: generate a string representation of this `_PropertyValue`
    __repr__ = lambda self: "Value(%s)%s" % ((
          '"%s"' % self.data) if isinstance(self.data, basestring)
              else repr(self.data), '*' if self.dirty else '')

  # = Internal Methods = #
  def __new__(cls, *args, **kwargs):

    """ Intercepts construction requests for directly Abstract model classes.

        Args:
        Kwargs:

        :raises:

        :returns: """

    if cls.__name__ == 'AbstractModel':  # prevent direct instantiation
      raise exceptions.AbstractConstructionFailure('AbstractModel')
    return super(AbstractModel, cls).__new__(cls, *args, **kwargs)

  # util: generate a string representation of this entity
  __repr__ = lambda self: "%s(%s, %s)" % (
      self.__kind__, self.__key__,
        ', '.join(['='.join([k, str(self.__data__.get(k, None))])
                   for k in self.__lookup__]))

  #__str__ = __unicode__ = __repr__  # map repr to str and unicode

  def __setattr__(self, name, value, exception=exceptions.InvalidAttribute):

    """ Attribute write override.

        :param name:
        :param value:
        :param exception:

        :raises:

        :returns: """

    # internal properties, data properties and `key`
    # can be written to after construction
    if name.startswith('__') or name in self.__lookup__ or name == 'key':
      # delegate upwards for write
      return super(AbstractModel, self).__setattr__(name, value)
    raise exception('set', name, self.kind())

  def __getitem__(self, name):

    """ Item getter support.

        :param name:

        :returns: """

    if name not in self.__lookup__:
      # only data properties are exposed via `__getitem__`
      raise exceptions.InvalidItem('get', name, self.kind())
    return getattr(self, name)  # proxy to attribute API

  # util: support for python's item API
  __setitem__ = lambda self, item, value: (
    self.__setattr__(item, value, exceptions.InvalidItem))

  def __context__(self, _type=None, value=None, traceback=None):

    """ Context enter/exit - apply explicit mode.

        :param _type:
        :param value:
        :param traceback:

        :returns: """

    if traceback:  # pragma: no cover
      return False  # in the case of an exception in-context, bubble it up
    self.__explicit__ = (not self.__explicit__)  # toggle explicit status
    return self

  # util: alias context entry/exit to `__context__` toggle method
  __enter__ = __exit__ = __context__

  # util: proxy `len` to length of written data (also alias `__nonzero__`)
  __len__ = lambda self: len(self.__data__)
  __nonzero__ = __len__

  # util: `dirty` property flag, proxies to `_PropertyValue`(s) for dirtyness
  __dirty__ = property(lambda self: any(
    (dirty for value, dirty in self.__data__.itervalues())))

  # util: `persisted` property flag, indicates whether key has been persisted
  __persisted__ = property(lambda self: self.key.__persisted__)

  def __iter__(self):

    """ Allow models to be used as dict-like generators.

        :returns: """

    for name in self.__lookup__:
      value = self._get_value(name, default=Property._sentinel)

      # skip unset properties without a default, except in `explicit` mode
      if (value == Property._sentinel and (not self.__explicit__)):
        if self.__class__.__dict__[name]._default != Property._sentinel:
          # return a prop's default in `implicit` mode
          yield name, self.__class__.__dict__[name].default
        continue  # pragma: no cover
      yield name, value
    raise StopIteration()

  def _set_persisted(self, flag=False):

    """ Notify this entity that it has been persisted to storage.

        :param flag:

        :returns: """

    self.key.__persisted__ = True
    for name in self.__data__:  # iterate over set properties
      # set value to previous, with `False` dirty flag
      self._set_value(name, self._get_value(name,
                      default=Property._sentinel), False)
    return self

  def _get_value(self, name, default=None):

    """ Retrieve the value of a named property on this Entity.

        :param name:
        :param default:

        :raises:
        :returns: """

    if name:  # calling with no args gives all values in (name, value) form
      if name in self.__lookup__:
        value = self.__data__.get(name, Property._sentinel)
        if not value:
          if self.__explicit__ and value is Property._sentinel:
            return Property._sentinel  # return EMPTY sentinel in explicit mode
          if callable(default):  # handle callable defaults
            return default(self)  # pragma: no cover
          return default  # return default value passed in
        return value.data  # return property value
      raise exceptions.InvalidAttribute('get', name, self.kind())
    return [(i, getattr(self, i)) for i in self.__lookup__]

  def _set_value(self, name, value=EMPTY, _dirty=True):

    """ Set (or reset) the value of a named property on this Entity.

        :param name:
        :param value:
        :param _dirty:

        :raises:
        :returns: """

    if not name: return self  # empty strings or dicts or iterables return self

    if isinstance(name, (list, dict)):
      if isinstance(name, dict):
        name = name.items()  # convert dict to list of tuples
      # filter out flags from caller
      return [self._set_value(k, i, _dirty=_dirty) for k, i in name if (
                k not in ('key', '_persisted'))]

    # allow a tuple of (name, value), for use in map/filter/etc
    if isinstance(name, tuple):  # pragma: no cover
      name, value = name

    if name == 'key':  # if it's a key, set through _set_key
      return self._set_key(value).owner  # returns `self` :)

    if name in self.__lookup__:  # check property lookup

      prop = self.__class__[name]  # extract property

      # inflate ISO-formatted datetimes
      if prop.basetype in (datetime.date, datetime.datetime) and not (
          isinstance(value, (datetime.date, datetime.datetime))):
        if isinstance(value, basestring):  # pragma: no cover
          # try to inflate from ISO
          value = dtparser.parse(value)
          if prop.basetype is datetime.date:
            value = value.date

        elif isinstance(value, (int, float)):  # pragma: no cover
          # try to inflate from timestamp
          value = datetime.datetime.fromtimestamp(value) if (
            prop.basetype is datetime.datetime) else (
            datetime.date.fromtimestamp(value))

      # if it's a valid property, create a namedtuple value placeholder
      self.__data__[name] = self.__class__._PropertyValue(value, _dirty)
      return self
    raise exceptions.InvalidAttribute('set', name, self.kind())

  def _set_key(self, value=None, **kwargs):

    """ Set this Entity's key manually.

        kwargs

        :param value:

        :raises:
        :returns: """

    _valid_key_classes = (self.__class__.__keyclass__, tuple, basestring)

    # cannot provide both a value and formats
    if value and kwargs:
      raise exceptions.MultipleKeyValues(self.kind(), value, kwargs)

    if issubclass(self, Vertex) and (
        isinstance(value, Key) and not isinstance(value, VertexKey)):
      value = VertexKey.from_urlsafe(value.urlsafe())  # pragma: no cover
    elif issubclass(self, Edge) and (
        isinstance(value, Key) and not isinstance(value, EdgeKey)):
      value = EdgeKey.from_urlsafe(value.urlsafe())  # pragma: no cover

    # for a literal key value
    if value is not None:
      if not isinstance(value, _valid_key_classes):  # filter out invalid keys
        raise exceptions.InvalidKey(*(
          type(value),
          value,
          self.__class__.__keyclass__.__name__))

      # set local key from result of dict->get(<formatter>)->__call__(<value>)
      self.__key__ = {

        # return keys directly
        self.__class__.__keyclass__: lambda x: x,

        # pass tuples through `from_raw`
        tuple: self.__class__.__keyclass__.from_raw,

        # pass strings through `from_urlsafe`
        basestring: self.__class__.__keyclass__.from_urlsafe

      }.get(type(value), lambda x: x)(value)._set_internal('owner', self)

      return self.__key__  # return key

    if kwargs:  # filter out multiple formats
      formatter, value = kwargs.items()[0]
      if len(kwargs) > 1:  # disallow multiple format kwargs
        raise exceptions.MultipleKeyFormats(', '.join(kwargs.keys()))

      # resolve key converter, if any, set owner, and `__key__`, and return
      self.__key__ = {

        # for raw, pass through `from_raw`
        'raw': self.__class__.__keyclass__.from_raw,

        # for strings, pass through `from_urlsafe`
        'urlsafe': self.__class__.__keyclass__.from_urlsafe,

        # by default it's a constructed key
        'constructed': lambda x: x

      }.get(formatter, lambda x: x)(value)._set_internal('owner', self)
      return self.__key__

    # except in the case of a null value and no formatter args
    raise exceptions.UndefinedKey(value, kwargs)

  ## = Property Bindings  = ##
  key = property(lambda self: self.__key__, _set_key)  # bind model key


## == Concrete Classes == ##

## Key
class Key(AbstractKey):

  """ Concrete Key class.
      DOCSTRING """

  __separator__ = u':'  # separator for joined/encoded keys
  __schema__ = (_DEFAULT_KEY_SCHEMA if not (
                _MULTITENANCY) else _MULTITENANT_KEY_SCHEMA)

  ## = Internal Methods = ##
  def __new__(cls, *parts, **formats):

    """ Constructs keys from various formats.

        args, kwargs

        :raises:
        :returns: """

    # extract 1st-provided format
    formatter, value = (
      formats.items()[0] if formats else ('__constructed__', None))

    # disallow multiple key formats
    try:
      _fmtgen = (i for i in formats if i != '_persisted')
      _fmtgen.next(), _fmtgen.next()
      raise exceptions.MultipleKeyFormats(', '.join(formats.keys()))
    except StopIteration:
      pass

    if not parts and formats.get('raw'):
      return cls.from_raw(formats['raw'])
    if not parts and formats.get('urlsafe'):
      return cls.from_urlsafe(formats['urlsafe'])

    return super(AbstractKey, cls).__new__(cls, *parts, **formats)

  def __init__(self, *parts, **kwargs):

    """ Initialize this Key.

        args, kwargs

        :raises:
        :returns: """

    if kwargs.get('raw') or kwargs.get('urlsafe') and (
        self.kind and self.id): return  # if we're re-initializing, return

    if len(parts) > 1:  # normal case: it's a full/partially-spec'd key

      # it's a fully- or partially-spec'ed key
      if len(parts) <= len(self.__schema__):
        _parts_diff = (len(self.__schema__) - len(parts))
        _pluck_kind = lambda x: x.kind() if hasattr(x, 'kind') else x
        mapped = zip([i for i in reversed(self.__schema__)][_parts_diff:],
                         map(_pluck_kind, parts))

      else:
        # for some reason the schema falls short of our parts
        raise exceptions.KeySchemaMismatch(*(
              self.__class__.__name__,
              len(self.__schema__),
              str(self.__schema__)))

      _pluck = lambda x: (x[0], x[1].kind()) if isinstance(x[1], Model) else x
      for name, value in map(_pluck, mapped):
        setattr(self, name, value)  # set appropriate attribute via setter

    elif len(parts) == 1:  # special case: it's a kinded, empty key
      if hasattr(parts[0], 'kind'):
        parts = (parts[0].kind(),)  # quick ducktyping: is it a model?
      self.__kind__ = parts[0]

    # if we *know* this is an existing key, `_persisted` should be `true`
    self._set_internal('parent', kwargs.get('parent'))

    # also set kwarg-passed parent.
    self._set_internal('persisted', kwargs.get('_persisted'))

  def __setattr__(self, name, value):

    """ Block attribute overwrites.

        :param name:
        :param value:

        :raises:
        :returns: """

    if not name.startswith('__'):
      if name not in self.__schema__:
        raise exceptions.InvalidKeyAttributeWrite('create', name, self)
      if getattr(self, name) is not None:
        raise exceptions.InvalidKeyAttributeWrite('overwrite', name, self)
    return super(AbstractKey, self).__setattr__(name, value)


class VertexKey(Key):

    """ Key class for ``Vertex`` records. """


class EdgeKey(Key):

    """ Key class for ``Edge`` records. """


## Property
class Property(object):

  """ Concrete Property class.
      DOCSTRING """

  __metaclass__ = abc.ABCMeta  # enforce definition of `validate` for subclasses
  __slots__ = _PROPERTY_SLOTS  # setup slots for property options
  _sentinel = sentinel = EMPTY  # default sentinel for basetypes/values

  ## = Internal Methods = ##
  def __init__(self, name, basetype,
                           default=_sentinel,
                           required=False,
                           repeated=False,
                           indexed=True,
                           **options):

    """ Initialize this Property.

        kwargs

        :param name:
        :param basetype:
        :param default:
        :param required:
        :param repeated:
        :param indexed:

        :raises:
        :returns: """

    # shim choices in from enum
    if isinstance(basetype, type) and (
        issubclass(basetype, BidirectionalEnum)) and not (
          options.get('choices')):
      options['choices'] = [choice for (key, choice) in basetype]

    # copy locals specified above onto object properties of the same name,
    # specified in `self.__slots__`
    map(lambda args: setattr(self, *args), (
          zip(self.__slots__, (
            name, options, indexed, required, repeated, basetype, default))))

  ## = Descriptor Methods = ##
  def __get__(self, instance, owner):

    """ Descriptor attribute access.

        :param instance:
        :param owner:

        :raises:
        :returns: """

    if isinstance(instance, Model):  # proxy to internal entity method.

      is_empty = (
        instance._get_value(self.name, default=Property.sentinel) is (
          Property.sentinel))
      if (self.options.get('embedded') is True) and (
              isinstance(self.basetype, type)) and (
              issubclass(self.basetype, Model)):

        # if we're an embedded submodel and so far we're unset, initialize
        # the property first with an empty instance
        if is_empty and not instance.__explicit__:
          ## @TODO(sgammon): embedded models that know about their encapsulators
          instance._set_value(self.name, self.basetype(), False)

        elif is_empty and instance.__explicit__:
          # if we have an empty or missing submodel in explicit mode, return
          # value directly, which should be the empty sentinel
          return Property.sentinel  # pragma: no cover

      # grab value, returning special
      # a) property default or
      # b) sentinel if we're in explicit mode and it is unset
      if self.default != Property.sentinel:  # we have a set default
        value = instance._get_value(self.name, default=self.default)
      else:
        value = instance._get_value(self.name, default=Property.sentinel)

      return None if (not value and (value == Property.sentinel) and
                      instance.__explicit__ is False) else value

    # otherwise, class-level access is always the property in question
    return self

  def __set__(self, instance, value):

    """ Descriptor attribute write.

        :param instance:
        :param value:

        :raises:
        :returns: """

    if instance is not None:  # only allow data writes after instantiation
      # delegate to `AbstractModel._set_value`
      return instance._set_value(self.name, value)
    raise exceptions.InvalidAttributeWrite('set', self.name, self.kind)

  # delegate to `__set__` to clear the property
  __delete__ = lambda self, instance: instance.__set__(instance, None)

  def valid(self, instance):

    """ Validate the value of this property, if any.

        :param instance:

        :raises:
        :returns: """

    if self.__class__ != Property and hasattr(self, 'validate'):
      # check for subclass-defined validator to delegate validation to
      return self.validate(instance)  # pragma: no cover

    value = instance._get_value(self.name)  # retrieve value

    # check required-ness
    if value is None or value is self.sentinel:
      if self.required:
        raise exceptions.PropertyRequired(self.name, instance.kind())
      # empty value, non-required, all good :)
      if value is self._sentinel: return True

    if isinstance(value, (list, tuple, set, frozenset)):  # check multi-ness
      if not self.repeated:
        raise exceptions.PropertyNotRepeated(self.name, instance.kind())
    else:
      if self.repeated:
        raise exceptions.PropertyRepeated(self.name, instance.kind())
      value = (value,)  # make value iterable

    for v in value:  # check basetype

      # it validates if:
      # 1) the field is typeless, or
      # 2) the value is `None` or an instance of it's basetype
      # 3) the value is a submodel and complies with:
      #    - being a valid model and kind (if embedded)
      #    - being a valid key (if not embedded)
      # 4) the value is a `Key` and we are an `Edge` seeking a `VertexKey`
      if self.basetype is None or (
        (v is not self.sentinel) and isinstance(v, (
                                      self.basetype, type(None)))):
        continue
      if isinstance(self.basetype, type):  # pragma: no cover
        if issubclass(self.basetype, Model):
          if self.options.get('embedded') and isinstance(v, self.basetype):
            continue  # embedded & compliant model
          elif not self.options.get('embedded') and isinstance(v, Key):
            continue  # non-embedded & compliant key
        elif issubclass(self.basetype, BidirectionalEnum):
          if v in self.basetype:
            continue  # consider bidirectional enums

      raise exceptions.InvalidPropertyValue(*(
        self.name, instance.kind(), type(v).__name__, self.basetype.__name__))
    return True  # validation passed! :)

  def __repr__(self):

    """ Generate a string representation
      of this :py:class:`Property`.

      :returns: Stringified, human-readable
      value describing this :py:class:`Property`. """

    return "%s(%s, type=%s)" % (
      self.__class__.__name__, self.name, self._basetype.__name__)

  __unicode__ = __str__ = __repr__

  # util method to clone `Property` objects
  clone = lambda self: self.__class__(self.name, self._basetype, self._default,
                                      self._required, self._repeated,
                                      self._indexed, **self._options)

  # config accessors
  basetype, required, repeated, indexed, options = (
    property(lambda self: self._basetype),
    property(lambda self: self._required),
    property(lambda self: self._repeated),
    property(lambda self: self._indexed),
    property(lambda self: self._options))

  @property
  def default(self):

    """ Accessor that collapses callback-based default values or returns any
        explicitly-set default value for a ``Property`` object.

        :returns: Default value at ``self._default`` if explicitly set. If
          ``self._default`` is a callable, it is called with the local model
          as the first and only parameter, and expected to return a value
          suitable for storage. """

    if callable(self._default):
      return self._default(self)  # pragma: no cover
    return self._default

  ## == Query Overrides (Operators) == ##
  __sort__ = lambda self, direction: (  # internal sort spawn
    query.Sort(self, operator=(direction or query.Sort.ASCENDING)))
  __filter__ = lambda self, other, operator: (  # internal filter spawn
    query.Filter(self, other, operator=(operator or query.Filter.EQUALS)))

  ## == Sort Spawn == ##
  __pos__ = lambda self: (
    self.__sort__(query.Sort.ASCENDING))  # `+` operator
  __neg__ = lambda self: (
    self.__sort__(query.Sort.DESCENDING))  # `-` operator

  ## == Filter Spawn == ##
  __eq__ = lambda self, other: (
    self.__filter__(other, query.Filter.EQUALS))  # `==` operator
  __ne__ = lambda self, other: (
    self.__filter__(other, query.Filter.NOT_EQUALS))  # `!=` operator
  __gt__ = lambda self, other: (
    self.__filter__(other, query.Filter.GREATER_THAN))  # `>` operator
  __ge__ = lambda self, other: (
    self.__filter__(other, query.Filter.GREATER_THAN_EQUAL_TO))  # `>=` operator
  __lt__ = lambda self, other: (
    self.__filter__(other, query.Filter.LESS_THAN))  # `<` operator
  __le__ = lambda self, other: (
    self.__filter__(other, query.Filter.LESS_THAN_EQUAL_TO))  # `<=` operator


class Model(AbstractModel):

  """ Concrete Model class. """

  __keyclass__ = Key

  ## = Internal Methods = ##
  def __init__(self, **properties):

    """ Initialize this Model.

        args

        :raises:
        :returns: """

    # grab key / persisted flag, if any, and set explicit flag to `False`
    self.__explicit__, self.__initialized__ = False, True

    # initialize key, internals, and map any kwargs into data
    self.key, self.__data__ = (
      properties.get('key') or self.__keyclass__(self.kind(), _persisted=False),
      {})

    self._set_value(properties, _dirty=(not properties.get('_persisted')))

  ## = Class Methods = ##
  kind = classmethod(lambda cls: cls.__name__)


class Vertex(Model):

  """ Concrete Vertex class.
      DOCSTRING """

  __owner__, __keyclass__ = 'Vertex', VertexKey

  class __metaclass__(AbstractModel.__metaclass__):

    """ DOCSTRING """

    def __gt__(cls, target):

      """ Syntactic sugar for creating an on-the-fly :py:class:`Edge` subclass.

          Overrides the syntax:
            ``class Friends(Person > Person): [...]```

          :param other: Target :py:class:`Vertex` subclass to factory an
            :py:class:`Edge` to.

          :returns: Factoried :py:class:`Edge` subclass that represents a type
            connecting ``self`` to ``other``. """

      return cls.to(target, directed=False)

    def __lt__(cls, origin):

      """ Syntactic sugar for creating an on-the-fly :py:class:`Edge` subclass.

          Overrides the syntax:
            ``class Supporting(Organization < Person): [...]```

          :param other: Target :py:class:`Vertex` subclass to factory an
            :py:class:`Edge` to.

          :returns: Factoried :py:class:`Edge` subclass that represents a type
            connecting ``self`` to ``other``. """

      return origin.to(cls, directed=False)

    def __rshift__(cls, target):

      """ Syntactic sugar for creating an on-the-fly :py:class:`Edge` subclass.
          This syntactic extension spawns a **directed** ``Edge``.

          Overrides the syntax:
            ``class Comment(Person >> Post): [...]```

          :param other: Target :py:class:`Vertex` subclass to factory an
            :py:class:`Edge` to.

          :returns: Factoried :py:class:`Edge` subclass that represents a type
            connecting ``self`` to ``other``. """

      return cls.to(target, directed=True)

    def __lshift__(cls, origin):

      """ Syntactic sugar for creating an on-the-fly :py:class:`Edge` subclass.
          This syntactic extension spawns a **directed** ``Edge``.

          Overrides the syntax:
            ``class Contribution(Legislator << Contributor): [...]```

          :param other: Target :py:class:`Vertex` subclass to factory an
            :py:class:`Edge` to.

          :returns: Factoried :py:class:`Edge` subclass that represents a type
            connecting ``self`` to ``other``. """

      return origin.to(cls, directed=True)

  @classmethod
  def to(cls, *targets, **options):

    """ Syntactic sugar for creating an on-the-fly :py:class:`Edge` subclass.

        :param target: Target :py:class:`Vertex` subclass to factory
          an :py:class:`Edge` to.

        :param directed: Whether the target :py:class:`Edge` subclass should be
          considered directional in nature.

        :raises TypeError: In the case that a non-``Vertex`` is passed for
          ``target``.

        :returns: Factoried :py:class:`Edge` subclass that represents a type
          connecting ``self`` to ``other``. """

    # classes only plz
    for target in targets:
      if not issubclass(target, Vertex):  # pragma: no cover
        raise TypeError("Vertices can only point to other Vertices."
                        " An Edge from `%s` to `%s` was requested, but `%s` is"
                        " not a Vertex." % (cls, target, target))

    return Edge.__spec_class__(cls, targets, options.get('directed', False))


class EdgeSpec(object):

  """ Specifies the peering and directed-ness of an :py:class:`Edge`. """

  __slots__ = ('origin', 'peering', 'directed')

  def __new__(cls, name=None, bases=None, pmap=None):

    """  """

    # subtype construction generates ``Edge`` subclasses
    if name and bases and isinstance(pmap, dict):
      return type(name, (Edge,), dict(pmap or {}, __spec__=bases[0]))

    # different signature if we're not constructing (for clarity)
    origin, peering, directed = name, bases, pmap

    # allow construction as well (for use as a config symbol, essentially)
    return object.__new__(cls, origin, peering, directed)

  def __init__(self, origin, peering, directed):

    """ Initialize an ``EdgeSpec`` class, according to the specified ``peering``
        configuration.

        :param origin:

        :param peering: An ordered iterable of ``Vertex`` subclasses to connect
          via an ``Edge``, where index ``0`` is always considered the **origin**
          and the remaining items are targets that can be validly connected to
          **origin**.

        :param directed: Flag (``bool``) indicating that this ``Edge``
          represents a directional relationship between each
          ``origin -> target`` pair. """

    if not len(peering) > 0:  # pragma: no cover
      raise TypeError('Cannot specify an `Edge` without at'
                      ' least one target.')

    self.origin, self.peering, self.directed = (
      origin, peering[0] if len(peering) == 1 else peering, directed)

  def __repr__(self):  # pragma: no cover

    """ Generate a string representation for the relationship specified by this
        ``EdgeSpec`` class.

        :returns: String, like ``(origin <-> target, ...)`` (if undirected),
          otherwise ``(origin -> target, ...)`` (if directed). """

    return "(%s %s %s)" % (
      self.origin.kind(), '<->' if not self.directed else '->', (
        self.peering.kind() if issubclass(self.peering, Model) else ', '.join((
          k.kind() for k in self.peering))))


class Edge(Model):

  """ Concrete Edge class.
      DOCSTRING """

  __owner__, __keyclass__ = 'Edge', EdgeKey


  class __metaclass__(Model.__metaclass__):

    """ DOCSTRING """

    @classmethod
    def initialize(mcs, name, bases, pmap):

      """ DOCSTRING """

      if '__spec__' not in pmap or pmap['__spec__'].directed:
        pmap['source'] = Key
        pmap['target'] = Key, {'repeated': True}
      if '__spec__' not in pmap or (
           '__spec__' in pmap and not pmap['__spec__'].directed):
        pmap['peers'] = Key, {'indexed': True, 'repeated': True}
      return super(mcs, mcs).initialize(name, bases, pmap)

  __spec_class__ = EdgeSpec

  ## = Internal Methods = ##
  def __init__(self, source=None, *targets, **properties):

    """ Initialize this ``Edge`` with a ``Vertex`` ``source`` and ``target``
        pair.

        :param pair_or_source:
        :param maybe_target:
        :param **properties:

        :raises:
        :returns: """

    _key_from_model = lambda obj: obj.key if isinstance(obj, Vertex) else obj
    _keyify = lambda k: (
      VertexKey.from_urlsafe(k) if isinstance(k, basestring) else (
        _key_from_model(k)))


    # @TODO(sgammon): better validation on inflation (directed vs not, etc)

    if not source and not targets and 'peers' in properties:
      source, target = tuple(properties['peers'])
      targets = (target,)  # should be length of 1

      if len(targets) > 1:  # pragma: no cover
        raise TypeError('Undirected `Edge` got multiple targets,'
                        ' where only one was expected:'
                        ' "%s".' % targets)

    if not source and 'source' in properties:  # pragma: no cover
      source = properties['source']
    if not targets and 'targets' in properties:  # pragma: no cover
      targets = properties['targets']

    if (source is None or not targets) and not (
        properties.get('_persisted')):  # pragma: no cover
      raise TypeError('Constructing an `Edge` requires at least'
                      ' one `source` and `target`, or one pair'
                      ' (`source`, `target`).')

    if hasattr(self, '__spec__') and self.__spec__.directed:
      properties['source'], properties['target'] = (
        _keyify(source), tuple((_keyify(t) for t in targets)))
    else:
      properties['peers'] = tuple([_keyify(source)] + list((
                                        _keyify(target) for target in targets)))

    super(Edge, self).__init__(**properties)


# Module Globals
__abstract__ = (abstract,
                MetaFactory,
                AbstractKey,
                AbstractModel)

__concrete__ = (concrete,
                Property,
                KeyMixin,
                ModelMixin,
                Key,
                Model,
                Vertex,
                Edge)

# All modules
__all__ = ('concrete',
           'abstract',
           'MetaFactory',
           'AbstractKey',
           'AbstractModel',
           'query',
           'Property',
           'KeyMixin',
           'ModelMixin',
           'Key',
           'Model',
           'Vertex',
           'Edge',
           'adapter',
           'exceptions')



# -*- coding: utf-8 -*-

"""

  canteen
  ~~~~~~~

  a minimal web framework for the modern web

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

debug, __version__ = __debug__, (0, 3)

# stdlib
import __builtin__; export = None

# yay modules!
from . import core
from . import rpc
from . import util
from . import test
from . import base
from . import model
from . import logic
from . import runtime
from . import dispatch
from . import exceptions

# canteen :)
from .rpc import *
from .core import *
from .util import *
from .test import *
from .base import *
from .logic import *
from .model import *
from .runtime import *
from .dispatch import *
from .exceptions import *


__all__ = [export for export in globals() if (
  export not in __builtin__.__dict__ and (
    not export.startswith('__')))]  # export all the things!



# -*- coding: utf-8 -*-

"""

  uwsgi runtime
  ~~~~~~~~~~~~~

  integrates :py:mod:`canteen` with :py:mod:`uwsgi`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# core
from ..core import runtime

# util
from ..util import debug
from ..util import decorators

# logic
from ..logic.realtime import TERMINATE


try:  # pragma: no cover

  with runtime.Library('uwsgi', strict=True) as (library, uwsgi):

    # logger
    logging = debug.Logger('uWSGI')


    @decorators.bind('uwsgi')
    class uWSGI(runtime.Runtime):

      """ WIP """

      def callback(self, start_response):

        """  """

        def responder(status, headers):

          """  """

          try:
            return start_response(status, headers)
          except IOError:
            return

        return responder

      def handshake(self, key, origin=None):

        """ WIP """

        uwsgi.websocket_handshake(key, origin)

      def send(self, payload, binary=False):

        """ WIP """

        return (uwsgi.websocket_send if not binary else (
              uwsgi.websocket_send_binary))(payload)

      def receive(self, blocking=True):

        """ WIP """

        try:
          return uwsgi.websocket_recv_nb if not blocking else (
                    uwsgi.websocket_recv)()

        except IOError:
          return TERMINATE

      def close(self):

        """ WIP """


    uWSGI.set_precedence(True)  # we're running *inside* uWSGI guys

except ImportError: pass



# -*- coding: utf-8 -*-

"""

  HTTP cookie logic
  ~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import json

# core runtime
from canteen.base import logic
from canteen.core import runtime

# canteen utils
from canteen.util import decorators

# core session API
from ..session import SessionEngine


with runtime.Library('werkzeug', strict=True) as (library, werkzeug):

  # subimports
  securecookie = library.load('contrib.securecookie')


  @decorators.bind('http.cookies')
  class Cookies(logic.Logic):

    """  """

    __modes__ = {}  # modes for dealing with cookies

    @classmethod
    def add_mode(cls, name):

      """  """

      def _mode_adder(mode_klass):

        """  """

        cls.__modes__[name] = mode_klass
        cls.__default__ = cls.__modes__[name]
        return mode_klass

      return _mode_adder

    @classmethod
    def get_mode(cls, name):

      """  """

      return cls.__modes__[name] if name in cls.__modes__ else cls.__default__


    @SessionEngine.configure('cookies')
    class CookieSessions(SessionEngine):

      """  """

      def load(self, request, http):

        """  """

        # if there is a session cookie, load it...
        if self.config.get('key', 'canteen') in request.cookies:

          # resolve serializer
          _serializer = Cookies.get_mode(self.config.get('mode', 'json'))

          # unserialize the cookie
          session = _serializer.unserialize(*(
            request.cookies[(
              self.config.get('key', 'canteen'))], self.api.secret))

          if not session.new: return request.set_session(session, self)

        # set an explicitly-empty session (none was found)
        return request.set_session(None, self)

      def commit(self, request, response, session):

        """  """

        # resolve serializer
        _serializer, _key = (
          Cookies.get_mode(self.config.get('mode', 'json')),
          self.config.get('key', 'canteen')
        )

        # serialize the cookie
        serialized = _serializer({
          'uuid': session.id
        }, self.api.secret).serialize()

        # do we even need to write the cookie?
        if _key in request.cookies and (request.cookies[_key] == serialized):
          return  # the cookies are equal: don't need to re-set it

        # cookie parameters
        _params = {}

        # resolve expiration
        if 'max_age' not in self.config:
          if 'expires' not in self.config:
            _params['max_age'] = None  # expire on session close
          else:
            _params['expires'] = self.config['expires']
        else:
          _params['max_age'] = self.config['max_age']

        # copy-over cookie params
        _params.update({
          'path': self.config.get('path', '/'),
          'secure': self.config.get('secure', False),
          'domain': self.config.get('domain', request.host.split(':')[0]),
          'httponly': self.config.get('http_only', (
                        self.config.get('httponly', False)))
        })

        # write cookie into the response
        response.set_cookie(_key, serialized, **_params)


  @Cookies.add_mode('json')
  class JSONCookie(securecookie.SecureCookie):

    """  """

    class CookieSerializer(json.JSONEncoder):

      """  """

      @classmethod
      def dumps(cls, structure):

        """  """

        return cls().encode(structure)

      @classmethod
      def loads(cls, serialized):

        """  """

        return json.loads(serialized)

      def default(self, obj):

        """  """

        if isinstance(obj, Exception):
          if hasattr(obj, 'message'):
            return obj.message

        return json.JSONEncoder.default(self, obj)

    serialization_method = CookieSerializer


  with runtime.Library('flask') as (flib, flask):

    # load flask sessions
    flask_sessions = flib.load('sessions')


    class FlaskSessionBridge(JSONCookie, flask_sessions.SessionMixin):
        """  """

    # install our session bridge, forcing Flask to use JSON cookies
    flask_sessions.SecureCookieSessionInterface.session_class = (
      FlaskSessionBridge)



# -*- coding: utf-8 -*-

"""

  session logic
  ~~~~~~~~~~~~~

  provides base logic for handling and enforcing user sessions.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


# stdlib
import time, hashlib, abc
import random, string, operator

# core, base, util
from ..core import hooks
from ..base import logic
from ..util import config
from ..util import decorators
from .. import model as models


## Globals
_BUILTIN_SESSION_PROPERTIES = frozenset((
  'seen',
  'data',
  'csrf',
  'established',
  'seen',
  'agent',
  'client',
  'tombstoned'))


class ClientSession(models.Model):

  """ Canteen model for a user or HTTP client session. Tracks things like a
      session ID, last-seen-time, and when the session was established. """

  seen = int  # when this session was last seen
  data = dict  # attached session state data
  csrf = basestring  # token for use in a CSRF
  agent = basestring  # useragent string last seen from this session
  client = basestring  # IP address last seen from this session
  tombstoned = bool, {'default': False}  # flag for destroying session by force
  established = int, {'required': True}  # timestamp for session establish


class Session(object):

  """ Thin object representing a user session. Used at runtime until it needs
      to be stored, which then spawns a model instance (usually a
      ``ClientSession``). """

  __slots__ = ('__id__', '__session__')

  def __init__(self, key=None,
                     model=ClientSession,
                     **kwargs):

    """ Initialize this ``Session`` object with an ID and session model.

        :param key: ``str`` key to attach to the session model and pass back-
          and-forth as the user's session ID.

        :param model: Model class to use for spawning the new session and
          persisting it.

        :param **kwargs: Extra keyword arguments to pass to ``model``'s
          constructor. """

    if isinstance(model, type):
      # set established timestamp if it's not there already
      # (keep ``seen`` up to date too)
      if 'established' not in kwargs:
        kwargs['established'] = kwargs['seen'] = int(time.time())

      if 'data' not in kwargs: kwargs['data'] = {}

      for k, v in kwargs.iteritems():
        if k not in _BUILTIN_SESSION_PROPERTIES:
          kwargs['data'][k] = v

      # it's a class
      key = Session.make_key(key, model)
      self.__id__, self.__session__ = (key.id,
                                       model(key=key, **kwargs))

    elif not kwargs:
      # it's an instance, set the ID and session
      self.__session__, self.__id__ = model, id

    else:
      raise RuntimeError('Cannot specify a session model'
                         ' instance and also additional kwargs.')

  # noinspection PyMethodParameters
  @decorators.classproperty
  def config(cls):

    """  """

    # @TODO(sgammon): convert to decorator
    return config.Config().get('Sessions', {'debug': True})

  ## == Accessors == ##
  id = property(lambda self: self.__id__)
  data = property(lambda self: self.__session__.data)
  csrf = property(lambda self: self.__session__.csrf or (
    setattr(self.__session__, 'csrf', self.generate_token()) or (
      self.__session__.csrf)))

  ## == Get/Set == ##
  def set(self, key, value, exception=Exception):

    """  """

    try:
      return setattr(self.data, key, value) or self
    except KeyError:
      if exception is not Exception:
          raise exception('Could not write to'
                          ' session item "%s".' % key)

  def get(self, key, default=None, exception=Exception):

    """  """

    if key in self.data: return self.data[key]
    if default: return default
    if exception is not Exception: raise exception('Could not resolve session'
                                                   ' data item "%s".' % key)

  # item protocol
  __getitem__ = lambda self, key: (
    self.get(key, exception=KeyError))

  __setitem__ = lambda self, key, value: (
    self.set(key, value, exception=KeyError))

  def __contains__(self, key):

    """  """

    return key in self.__session__.data

  ## == Reset == ##
  def reset(self, save=False, adapter=None):

    """  """

    # tombstone and clear CSRF
    self.__session__.csrf, self.__session__.tombstoned = None, True
    if save: self.save(adapter)
    return

  def reset_csrf(self, save=False, adapter=None):

    """  """

    # clear the current CSRF
    self.__session__.csrf = None

    # generate a new one and return, possibly forcing a save
    new_csrf = self.csrf
    if save: self.save(adapter)
    return new_csrf

  ## == Save/Load == ##
  def save(self, environ, adapter=None):

    """  """

    if 'REMOTE_ADDR' in environ:
      self.__session__.client = environ.get('REMOTE_ADDR')

    if 'HTTP_USER_AGENT' in environ:
      self.__session__.agent = environ.get('HTTP_USER_AGENT')

    if self.config.get('storage', {}).get('enable'):
      return self.__session__.put(adapter=adapter)

  @classmethod
  def load(cls, id, model=ClientSession, strict=False, data=None):

    """  """

    # manufacture our own session, by loading the model
    #_session = model.get(Session.make_key(id, model))
    _session = None  # @TODO(sgammon): fix this
    if _session: return cls(id, _session)
    if strict: return False  # possibly fail stuff hardcore
    return cls(id, data=dict(data.iteritems()))  # otherwise make a new one

  ## == Session IDs == ##
  @staticmethod
  def generate_token(salt=''):

    """  """

    return Session.config.get('hash', hashlib.sha256)(
      salt + reduce(operator.add, (
        random.choice(string.printable) for x in xrange(32)))
    ).hexdigest()

  @staticmethod
  def make_key(id=None, model=ClientSession):

    """  """

    return models.Key(model, id or (
      Session.generate_token(Session.config.get('salt', ''))))


class SessionEngine(object):

  """  """

  ## == Internals == ##
  __label__, __metaclass__ = None, abc.ABCMeta
  __api__, __path__, __config__ = None, None, {}

  def __init__(self, name, config, api):

    """  """

    self.__path__, self.__config__, self.__api__ = name, config, api

  ## == Accessors == ##
  api = property(lambda self: self.__api__)
  config = property(lambda self: self.__config__.get(self.__path__))
  session_config = property(lambda self: self.__config__)

  @staticmethod
  def configure(name, **config):

    """  """

    def add_engine(klass):

      """  """

      klass.__label__ = name
      Sessions.add_engine(name, klass, **config)
      return klass

    return add_engine

  ## == Abstract Methods == ##
  @abc.abstractmethod
  def load(self, context):

    """  """

    raise NotImplementedError('Method `SessionEngine.load` is abstract.')

  @abc.abstractmethod
  def commit(self, context, session):

    """  """

    raise NotImplementedError('Method `SessionEngine.commit` is abstract.')


@decorators.bind('sessions')
class Sessions(logic.Logic):

  """  """

  ## == Internals == ##
  __salt__ = None  # the secret value to prepend to the cookie before hashing
  __secret__ = None  # the secret value to use in cookie checksumming
  __engines__ = {}  # mapping of names to engines that are supported
  __algorithm__ = None  # hash algorithm to share between engines

  @decorators.classproperty
  def config(cls):

    """  """

    return config.Config().get('Sessions', {'debug': True})

  @decorators.classproperty
  def salt(cls):

    """  """

    if not cls.__salt__:
      cls.__salt__ = cls.config.get('salt')
      if not cls.__salt__:
        cls.__salt__ = Session.generate_token()
    return cls.__salt__

  @decorators.classproperty
  def secret(cls):

    """  """

    if not cls.__secret__:
      cls.__secret__ = cls.config.get('secret')
      if not cls.__secret__:
        cls.__secret__ = (Session.generate_token() + Session.generate_token())
    return cls.__secret__

  @decorators.classproperty
  def engines(cls):

    """  """

    for engine in cls.__engines__.iteritems():
      yield engine

  @classmethod
  def add_engine(cls, name, engine, **config):

    """  """

    cls.__engines__[name] = (engine, config)
    return cls

  @classmethod
  def get_engine(cls, name=None, context=None):

    """  """

    _CONTEXT, _context_cfg = (
      (False, {}) if not context else (
        True, config.Config().get(context, {}).get('sessions', {})))

    # try looking in config if no engine is specified
    if not name: name = _context_cfg.get('engine', 'cookies')

    # look for an explicit engine by name if requested
    if name in cls.__engines__:

      _engine, _engine_config = cls.__engines__[name]

      # build config, allowing overrides
      _config = {}
      if _CONTEXT:
        _config.update(config.Config().get('Sessions', {}))
        _config.update(_context_cfg)
      _config.update(_engine_config)

      return _engine(name, _config, cls)
    raise RuntimeError('No such session runtime: "%s".' % name)

  ## == Injected Methods == ##
  @decorators.bind('reset')
  def reset(self, redirect=None, save=True, engine=None):

    """  """

    pass

  @decorators.bind('establish', wrap=hooks.HookResponder('match', context=(
    'environ', 'endpoint', 'arguments', 'request', 'http')))
  def establish(self, environ, endpoint, arguments, request, http):

    """  """

    if request.session:  # are sessions enabled?

      session, engine = request.session  # extract session + engine

      # engine is loaded, but no session
      if not session and self.config.get('always_establish', True):
        return request.set_session(Session(), engine)

  @decorators.bind('load', wrap=hooks.HookResponder(*(
    'request', 'message'), context=('request', 'http')))
  def load(self, request, http):

    """  """

    if http:  # HTTP sessions

      assert request, "must have a request to load a session"

      session_cfg = http.config.get('sessions', {'enable': False})
      if session_cfg.get('enable', True):

        # find us an engine, yo, and default to cookie-based sessions (safest)
        engine = self.get_engine(name=session_cfg.get('engine', 'cookies'),
                                 context='http')

        engine.load(request=request, http=http)

  @decorators.bind('commit', wrap=hooks.HookResponder('response', context=(
    'status', 'headers', 'request', 'http', 'response')))
  def commit(self, status, headers, request, http, response):

    """  """

    if response:  # we can only apply sessions to full responses
      if request.session:  # are sessions enabled?

        session, engine = request.session  # extract engine and session

        # defer to engine to commit
        engine.commit(request=request, response=response, session=session)

  @decorators.bind('save', wrap=hooks.HookResponder('complete', context=(
                                    'response', 'request', 'http', 'environ')))
  def save(self, response, request, http, environ):

    """  """

    if request.session and response:  # are sessions enabled?

      session, engine = request.session  # extract engine and session

      # @TODO(sgammon): support custom adapters here

      # save the session via backend, along with request context
      session.save(environ, adapter=None)



# -*- coding: utf-8 -*-

"""

  tests
  ~~~~~

  utilities for providing unittest functionality.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from __future__ import print_function


if __debug__:

  # stdlib
  import os
  import sys
  import unittest
  import itertools

  # internals
  from .core import meta
  from .core import Library
  from .util import config
  from .util import ObjectProxy

  # dispatch tools
  from .dispatch import spawn

  # HTTP logic
  from .logic.http import semantics


  ## Globals
  combine = lambda *t: tuple(itertools.chain(*t))
  _get_app = lambda s, k: k.get('app', s.dispatch_endpoint())
  _dispatch = lambda m, s, a, k: (
      s.dispatch(*combine((_get_app(s, k), m), a), **k))


  class BaseTest(unittest.TestCase):

    """ Provides base code testing functionality, both for framework testing
        and testing of applications written with Canteen. """

    __appconfig__ = None  # class-level assignment of app config state
    __orig_testing__ = None  # original CANTEEN_TESTING flag from environ

    @classmethod
    def setUpClass(cls):  # pragma: no cover

      """ Test class setup hook, in this case used for forcing the env var
          ``CANTEEN_TESTING`` to exist.

          :returns: ``None``. """

      if 'CANTEEN_TESTING' not in os.environ:  # pragma: no cover
        cls.__orig_testing__ = False  # no original flag
        os.environ['CANTEEN_TESTING'] = 'on'
      super(BaseTest, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):  # pragma: no cover

      """ Test class teardown hook, in this case used to remove the env var
          ``CANTEEN_TESTING`` if it was forced into existence.

          :returns: ``None``. """

      if not cls.__orig_testing__:  # pragma: no cover
        del os.environ['CANTEEN_TESTING']

    @classmethod
    def set_config(cls, target):

      """ Set app configuration to a testing-compatible active set.

          :param target: :py:class:`canteen.util.config.Config` instance that
            contains app/framework configuration to be used during testing.

          :returns: ``cls``, for chainability. """

      return setattr(cls, '__appconfig__', target) or cls  # pragma: no cover


    with Library('werkzeug', strict=True) as (library, werkzeug):

      __werkzeug__ = ObjectProxy({
        'test': library.load('test'),
        'testapp': library.load('testapp'),
        'wrappers': library.load('wrappers')
      })

      def dispatch_endpoint(self):  # pragma: no cover

        """ Internal function to spawn a throwaway app instance for issuing a
            one-shot dispatch.

            :returns: Instance of :py:class:`canteen.core.Runtime`, for
              whichever runtime is active for the target app. """

        return spawn(None, config=config.Config(**self.__appconfig__))

      def _spawn_client(self, wsgi_target=None):  # pragma: no cover

        """ Internal function to spawn a throwaway :py:mod:`werkzeug`-based HTTP
            :py:class:`werkzeug.test.Client` instance, for the purpose of
            executing WSGI dispatch during testing.

            :param wsgi_target: Target WSGI application to dispatch against.
              Defaults to ``None``, indicating that a reference WSGI app (known
              to be good) should be used in its place.

            :returns: Prepared :py:class:`werkzeug.test.Client` instance, with
              the target ``wsgi_target`` wrapped and ready to dispatch. """

        return self.__werkzeug__.test.Client(*(
          wsgi_target or self.__werkzeug__.testapp.test_app,
          semantics.HTTPSemantics.HTTPResponse))

      def dispatch(self, app, method, *args, **kwargs):  # pragma: no cover

        """ Perform a WSGI dispatch against ``app``, with HTTP ``method`` and
            pass position ``args`` and keyword ``kwargs``.

            :param app: WSGI application to dispatch against.

            :param method: HTTP method to dispatch in target ``app``.

            :param *args: Positional arguments to pass to the Werkzeug test
              client dispatch method.

            :param **kwargs: Keyword arguments to pass to the Werkzeug test
              client dispatch method.

            :returns: Result of dispatching ``method`` against ``app`` via
              WSGI. """

        return getattr(self._spawn_client(app), method.lower())(*args, **kwargs)

      # HTTP method aliases
      GET = lambda self, *a, **k: _dispatch('GET', self, a, k)
      PUT = lambda self, *a, **k: _dispatch('PUT', self, a, k)
      POST = lambda self, *a, **k: _dispatch('POST', self, a, k)
      HEAD = lambda self, *a, **k: _dispatch('HEAD', self, a, k)
      TRACE = lambda self, *a, **k: _dispatch('TRACE', self, a, k)
      PATCH = lambda self, *a, **k: _dispatch('PATCH', self, a, k)
      PURGE = lambda self, *a, **k: _dispatch('PURGE', self, a, k)
      DELETE = lambda self, *a, **k: _dispatch('DELETE', self, a, k)
      OPTIONS = lambda self, *a, **k: _dispatch('OPTIONS', self, a, k)
      CONNECT = lambda self, *a, **k: _dispatch('CONNECT', self, a, k)


  class AppTest(BaseTest):

    """ Provides an extension point for application developers to write tests
        against applications written for ``Canteen``. """

    __root__, __owner__, __metaclass__ = True, 'AppTest', meta.Proxy.Registry


  class FrameworkTest(BaseTest):

    """ Provides an extension point for Canteen developers (on the framework
        itself) to create tests for builtin functionality. """

    __root__, __owner__, __metaclass__ = (
      True, 'FrameworkTest', meta.Proxy.Registry)


  def run(output=None,
          suites=None,
          scope=(AppTest, FrameworkTest),
          _format='text',
          verbosity=1, **kwargs):  # pragma: nocover

    """ Run a suite of test cases with an optional output plan, and optionally
        scoping to only framework or application tests.

         :param output:
         :param suites:
         :param scope:
         :param _format:
         :param verbosity:
         :param kwargs:

         :returns: """

    # fill testsuite with found testcases
    master_suite, loader = [], unittest.TestLoader()

    for bucket in scope:
      suite = unittest.TestSuite()

      for child in bucket.iter_children():
        suite.addTests(loader.loadTestsFromTestCase(child))

      master_suite.append(suite)

    if suites:
      for _suite in suites:
        master_suite.append(_suite)

    def filter_suite(_s):

      """ Filter whole testsuites from being run by this tool if they don't
          contain any substantive tests.

          :param _s: Testsuite to filter.

          :returns: ``True`` if the testsuite should be run, ``False``
            otherwise. """

      if not _s.countTestCases():
        return False
      return True

    _seen_tests = set()

    def merge_suite(left, right):

      """ Merge two testsuites' containing tests, into one testsuite.

          :param left: The first testsuite to merge from.
          :param right: The other testsuite to merge from.

          :returns: :py:class:`unittest.TestSuite` instance merged of all tests
            contained in ``left`` and ``right``. """

      _master = []
      for case in [test for test in left] + [test for test in right]:
        if isinstance(case, unittest.TestSuite):
          for _case in case:
            _master.append(_case)
            _seen_tests.add(_case)
          continue
        if case not in _seen_tests:
          _master.append(case)
          _seen_tests.add(case)
        continue
      return set(unittest.TestSuite(_master))

    master_suite = unittest.TestSuite(reduce(merge_suite, (
      filter(filter_suite, master_suite))))

    # allow for XML format
    if _format == 'xml':
      if output is None:
        output = ".develop/tests"
      try:
        # noinspection PyPackageRequirements
        import xmlrunner
        return xmlrunner.XMLTestRunner(output=output).run(master_suite)
      except ImportError:
        xmlrunner = False
        raise RuntimeError('Cannot generate XML output without `xmlrunner`.')
    runner = unittest.TextTestRunner(stream=output or sys.stdout,
                                     verbosity=verbosity, **kwargs)
    return runner.run(master_suite)

  def clirunner(arguments, root=None):  # pragma: nocover

    """ Discover and run known testsuites. Optionally scope to a certain
        ``root`` directory, or provide alternate output options.

        :param arguments: Command-line arguments.
        :param root: Root directory to run tests from.

        :returns: Nothing, as ``sys.exit`` is called from this tool. """

    output, _format = None, 'text'

    if not __debug__:
      raise RuntimeError('Cannot run tests with -O or -OO.')
    if not root:
      root = os.getcwd()

    if arguments:
      if len(arguments) > 2:
        print("Can only call with a maximum of 2 arguments:"
              " FORMAT and OUTPUT, or just FORMAT.")
        sys.exit(1)
      if len(arguments) == 2:
        _format, output = tuple(arguments)
      else:
        _format = arguments[0]

    discovered = None
    if root:
      loader = unittest.TestLoader()
      discovered = loader.discover(root)

    try:
      run(**{
        'output': output or (sys.stdout if _format is 'text' else None),
        'suites': discovered,
        'format': _format,
        'verbosity': 5 if 'TEST_VERBOSE' in os.environ else (
          0 if 'TEST_QUIET' in os.environ else 1)
      })
    except Exception as e:
      print(e)
      sys.exit(1)
    else:
      sys.exit(0)



# -*- coding: utf-8 -*-

"""

  model exceptions
  ~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


class Error(Exception):

  """ Base model error class. """

  def __init__(self, *args, **kwargs):

    """ Initialize model-layer exception.

        args/kwargs """

    self.fmt_args, self.fmt_kwargs = args, kwargs

  __repr__ = __str__ = __unicode__ = lambda self: (
    self.message % (self.fmt_kwargs if self.fmt_kwargs else self.fmt_args))


class AbstractConstructionFailure(Error, NotImplementedError):

  """  """

  message = "Cannot directly instantiate abstract class `%s`."


class AdapterException(Error, RuntimeError):

  """  """

  pass


class NoSupportedAdapters(AdapterException):

  """  """

  message = "No valid model adapters found."


class InvalidExplicitAdapter(AdapterException):

  """  """

  message = ("Requested model adapter \"%s\" could not be found"
             " or is not supported in this environment.")


class InvalidKey(TypeError):

  """  """

  message = ("Cannot set model key to invalid type \"%s\" (for value \"%s\")."
             " Expected `basestring`, `tuple` or `%s`.")


class UndefinedKey(InvalidKey):

  """  """

  message = ("Could not operate on undefined key"
             " (value: \"%s\", kwargs: \"%s\").")


class MultipleKeyValues(TypeError):

  """  """

  message = ("Cannot merge multiple key values/formats in `%s._set_key`."
             " (got: value(%s), formats(%s)).")


class MultipleKeyFormats(TypeError):

  """  """

  message = "Cannot provide multiple formats to `_set_key` (got: \"%s\")."


class PersistedKey(AttributeError):

  """  """

  message = "Cannot set property \"%s\" of an already-persisted key."


class InvalidAttributeWrite(AttributeError):

  """  """

  message = "Cannot %s property \"%s\" of model \"%s\" before instantiation."


class InvalidKeyAttributeWrite(AttributeError):

  """  """

  message = "Cannot %s property \"%s\" of key \"%s\" before instantiation."


class InvalidAttribute(AttributeError):

  """  """

  message = "Cannot %s nonexistent data property \"%s\" of model class \"%s\"."


class InvalidItem(KeyError):

  """  """

  message = "Cannot %s nonexistent data item \"%s\" of model class \"%s\"."


class KeySchemaMismatch(InvalidKey):

  """  """

  message = ("Key type \"%s\" takes a maximum of %s positional arguments"
             " to populate the format \"%s\".")


class ValidationError(ValueError):

  """  """

  pass


class PropertyPolicyViolation(ValidationError):

  """  """

  pass


class PropertyBasetypeViolation(ValidationError):

  """  """

  pass


class PropertyRequired(PropertyPolicyViolation):

  """  """

  message = ("Property \"%s\" of Model class \"%s\" is marked as `required`,"
             " but was left unset.")


class PropertyRepeated(PropertyPolicyViolation):

  """  """

  message = ("Property \"%s\" of Model class \"%s\" is marked as iterable,"
             " and cannot accept non-iterable values.")


class PropertyNotRepeated(PropertyPolicyViolation):

  """  """

  message = ("Property \"%s\" of Model class \"%s\" is not marked as repeated,"
             " and cannot accept iterable values.")


class InvalidPropertyValue(PropertyBasetypeViolation):

  """  """

  message = ("Property \"%s\" of Model class \"%s\" cannot accept value of type"
             " \"%s\" (was expecting type \"%s\").")



# -*- coding: utf-8 -*-

"""

  cli tests
  ~~~~~~~~~

  tests for canteen's data structures utilities.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen
from canteen.util import cli
from canteen import FrameworkTest


class CLIToolsTests(FrameworkTest):

  """ Tests for `cli.Tool` """

  def test_construct(self):

    """ Test construction of a simple CLI tool """

    class Sample(cli.Tool):

      """ sample CLI tool """

      def execute(arguments):

          """ execution flow """

    assert isinstance(Sample.__dict__['execute'], staticmethod), (
      "by default tool execution methods should be static")
    return Sample

  def test_construct_subtool(self):

    """ Test construction of a CLI tool with subtools """

    class Sample(cli.Tool):

      """ sample CLI tool """

      class Subsample(cli.Tool):

        """ sub-sample CLI tool """

        @classmethod
        def execute(cls, arguments):

            """ sample """

    assert isinstance(Sample.Subsample.__dict__['execute'], classmethod), (
      "classmethods should be allowed as tool execution flows, instead got"
      " '%s'" % repr(Sample.Subsample.execute))
    return Sample

  def test_construct_arguments(self):

    """ Test construction of a CLI tool with arguments without short options """

    class Sample(cli.Tool):

      """ sample CLI tool """

      arguments = (
        ('--debug', {'action': 'store_true'}),)

      class Subsample(cli.Tool):

        """ sub-sample CLI tool """

        def execute(arguments):

            """ sample """

    return Sample

  def test_construct_arguments_with_short(self):

    """ Test construction of a CLI tool with arguments with short options """

    class Sample(cli.Tool):

      """ sample CLI tool """

      arguments = (
        ('--debug', '-d', {'action': 'store_true'}),)

      class Subsample(cli.Tool):

        """ sub-sample CLI tool """

        def execute(arguments):

            """ sample """

    return Sample

  def test_initialize_clitool(self):

    """ Test initializing a CLI tool in various contexts without safe mode """

    self.test_construct()(autorun=False, safe=False)
    self.test_construct_subtool()(autorun=False, safe=False)
    self.test_construct_arguments()(autorun=False, safe=False)
    self.test_construct_arguments_with_short()(autorun=False, safe=False)

  def test_initialize_clitool_safe(self):

    """ Test initializing a CLI tool in various contexts with safe mode """

    self.test_construct()(autorun=False, safe=True)
    self.test_construct_subtool()(autorun=False, safe=True)
    self.test_construct_arguments()(autorun=False, safe=True)
    self.test_construct_arguments_with_short()(autorun=False, safe=True)



# -*- coding: utf-8 -*-

"""

  core
  ~~~~

  classes, utilities and stuff that is core to the proper operation of canteen.
  meta stuff, abstract stuff, and runtime utilities.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# submodules
from .meta import *
from .hooks import *
from .runtime import *
from .injection import *


__all__ = ('meta',
           'hooks',
           'runtime',
           'injection')



# -*- coding: utf-8 -*-

"""

  werkzeug runtime
  ~~~~~~~~~~~~~~~~

  runs :py:mod:`canteen`-based apps on pocoo's excellent WSGI
  library, :py:mod:`werkzeug`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib & core
import os
from ..util import struct
from ..core import runtime


with runtime.Library('werkzeug', strict=True) as (library, werkzeug):

  # WSGI devserver
  serving, err = library.load('serving'), library.load('exceptions')

  http_exceptions = {
    'BadRequest': err.BadRequest,  # 400
    'Unauthorized': err.Unauthorized,  # 401
    'Forbidden': err.Forbidden,  # 403
    'NotFound': err.NotFound,  # 404
    'MethodNotAllowed': err.MethodNotAllowed,  # 405
    'NotAcceptable': err.NotAcceptable,  # 406
    'RequestTimeout': err.RequestTimeout,  # 408
    'Conflict': err.Conflict,  # 409
    'Gone': err.Gone,  # 410
    'LengthRequired': err.LengthRequired,  # 411
    'PreconditionFailed': err.PreconditionFailed,  # 412
    'RequestEntityTooLarge': err.RequestEntityTooLarge,  # 413
    'RequestURITooLarge': err.RequestURITooLarge,  # 414
    'UnsupportedMediaType': err.UnsupportedMediaType,  # 415
    'RequestedRangeNotSatisfiable': err.RequestedRangeNotSatisfiable,  # 416
    'ExpectationFailed': err.ExpectationFailed,  # 417
    'ImATeapot': err.ImATeapot,  # 418
    'PreconditionRequired': err.PreconditionRequired,  # 428
    'TooManyRequests': err.TooManyRequests,  # 429
    'RequestHeaderFieldsTooLarge': err.RequestHeaderFieldsTooLarge,  # 431
    'InternalServerError': err.InternalServerError,  # 500
    'NotImplemented': err.NotImplemented,  # 501
    'ServiceUnavailable': err.ServiceUnavailable,  # 502
    'ClientDisconnected': err.ClientDisconnected,
    'SecurityError': err.SecurityError}


  class Werkzeug(runtime.Runtime):

    """  """

    base_exception = err.HTTPException

    exceptions = struct.ObjectProxy(http_exceptions)

    def bind(self, interface, address):  # pragma: no cover

      """  """

      paths = {}

      # resolve static asset paths
      if 'assets' in self.config.app.get('paths', {}):
        if isinstance(self.config.app['paths'].get('assets'), dict):
          paths.update(dict((
            (k, v) for k, v in self.config.app['paths']['assets'].iteritems())))

        paths.update({
          '/assets': self.config.app['paths']['assets'],
          '/favicon.ico': self.config.app['paths'].get('favicon', False) or (
            os.path.join(self.config.app['paths']['assets'], 'favicon.ico'))})

      # append any extra asset paths
      if self.config.assets.get('config', {}).get('extra_assets'):
        paths.update((
          dict(self.config.assets['config']['extra_assets'].itervalues())))

      # run via werkzeug's awesome `run_simple`
      return serving.run_simple(interface, address, self, **{
        'use_reloader': True,
        'use_debugger': True,
        'use_evalex': True,
        'extra_files': None,
        'reloader_interval': 1,
        'threaded': True,
        'processes': 1,
        'passthrough_errors': False,
        'ssl_context': None,
        'static_files': paths})



# -*- coding: utf-8 -*-

"""

  HTTP logic
  ~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# symbols
from .agent import UserAgent
from .cookies import Cookies
from .semantics import HTTPSemantics, url


# exports
__all__ = ('url',
           'Cookies',
           'UserAgent',
           'HTTPSemantics')



# -*- coding: utf-8 -*-

"""

  dispatch
  ~~~~~~~~

  WSGI dispatch entrypoint. INCEPTION.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import os


## Globals
app = None


def spawn(app,
          config=None):  # pragma: no cover

  """ Spawn a Canteen app, suitable for dispatch
      as a WSGI application.

      :param app: Canteen application to be spawned,
        usually passed as a Python module.

      :param config: Application configuration, in
        the form of a ``canteen.util.Config`` instance
        wrapping a dictionary of application config.

      :returns: Instance of ``canteen.Runtime`` that
        can be dispatched via WSGI and wraps the target
        ``app`` object.  """

  # canteen core & util
  from canteen.core import runtime
  from canteen.util import config as cfg

  if not config: config = cfg.Config()
  return runtime.Runtime.spawn(app).configure(config)


# @TODO(sgammon): wtf is root
def run(app=None,
        interface='127.0.0.1',
        port=8080,
        dev=True,
        config=None):  # pragma: no cover

  """ Run a lightweight development server via the
      currently-active runtime. Suitable for use
      locally, with no required parameters at all.

      :param app: Canteen application to be spawned,
        usually passed as a Python module.

      :param root: Unused. No fucking clue what this
        is but I'd guess it's the root filepath to the
        application. I hope it's not that, though,
        because that would break App Engine.

      :param interface: Network interface that should
        be bound to for the resulting lightweight HTTP
        server. Defaults to ``127.0.0.1``.

      :param port: Integer port number that should be
        bound to for the resulting lightweight HTTP
        server. Defaults to ``8080``.

      :param dev: Boolean flag indicating whether we
        should be running in debug mode or not. Controls
        various things like log output. Defaults to
        ``True`` as this method is only meant to be an
        easy way to put up a dev server.

      :returns: Nothing useful, as this blocks to
        serve requests forever and ever. """

  if 'CANTEEN_TESTING' in os.environ and os.environ['CANTEEN_TESTING'] in (
    'yep', '1', 'sure', 'ofcourse', 'whynot', 'yes', 'on'):
    return spawn(app, config or {})  # pragma: no cover
  return spawn(app, config or {}).serve(interface, port)



# -*- coding: utf-8 -*-

"""

  RPC exceptions
  ~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from . import ServerException
from . import ClientException
from . import Exception as Error


class InternalRPCException(Error):

    """ Base class for all errors in service handlers module. """


class ServiceConfigurationError(InternalRPCException):

    """ When service configuration is incorrect. """


class RequestError(ClientException):

    """ Error occurred when building request. """


class ResponseError(ServerException):

    """ Error occurred when building response. """



# -*- coding: utf-8 -*-

"""

  HTTP agent logic
  ~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen core & util
from canteen.base import logic
from canteen.util import struct
from canteen.util import decorators


class Vendor(struct.BidirectionalEnum):

  """ Enumerated common vendors that can be found in an HTTP client or browser's
      ``User-Agent`` string. """

  GOOGLE = 0x0  # Chrome
  MOZILLA = 0x1  # Firefox
  MICROSOFT = 0x2  # IE
  OPERA = 0x3  # Opera
  APPLE = 0x4  # Apple
  OPEN = 0x5  # Open source
  OTHER = 0x6  # Anything else


class AgentInfo(object):

  """ Base class structure that removes object slots in favor of an extendable,
      fully-static object structure. Used for indivisual objects that retain or
      specify details about an HTTP client's ``User-Agent`` string. """

  __slots__ = tuple()

  def dump(self):

    """ Dump the local data carried by this ``AgentInfo`` object or subclass
        object.

        :returns: Dictionary (``dict``) of held ``key => value`` pairs. """

    return dict(((k, getattr(self, k, None)) for k in self.__slots__))

  def __repr__(self):

    """ Generate a pleasant string representation for this unit of
        ``AgentInfo``.

        :returns: Human-readable string representation of this object. """

    return "%s(%s)" % (
      self.__class__.__name__.replace('Agent', ''),
      ', '.join(
        ('='.join((
          i,
          str(getattr(self, i) if (
            hasattr(self, i)) else None))) for i in (
              self.__slots__) if not i.startswith('__'))))


class AgentVersion(AgentInfo):

  """ Holds parsed version information for a software HTTP client or browser,
      found while scanning the ``User-Agent`` header. """

  __slots__ = (
    'major',  # major browser version (the `3` in 3.0)
    'minor',  # minor browser version (the `1` in 3.1)
    'micro')  # micro browser version (the `5` in 3.1.5)

  def __init__(self, major, minor=None, micro=None):

    """ Initialize this version info container.

        :param major: Major version.
        :param minor: Minor version (optional, defaults to ``None``).
        :param micro: Micro version (optional, defaults to ``None``). """

    self.major, self.minor, self.micro = major, minor, micro


class AgentOS(AgentInfo):

  """ Holds parsed operating system information for a software HTTP client or
      browser, found while scanning the ``User-Agent`` string. """

  __slots__ = (
    'name',  # `Mac OS X`, `Windows XP`, etc
    'vendor',  # vendor of the OS, from above
    'version')  # detected version of the OS

  def __init__(self, name, vendor, version):

    """ Initialize this OS info container.

        :param name: Name of the operating system running on the host described
          by the subject ``User-Agent`` string.

        :param vendor: Software vendor that produced the operating system
          running on the host described by the ``User-Agent`` string.

        :param version: Version information for the operating system running on
          the host described by the ``User-Agent`` string. """

    self.name, self.vendor, self.version = name, vendor, version

  @classmethod
  def scan(cls, request, user_agent, detected):

    """ Scan a target ``user_agent`` string, encapsulated by an HTTP
        ``request``, for information about an HTTP client or browser's active
        operating system.

        :param request: HTTP request that carries with it the ``User-Agent``
          header in question.

        :param user_agent: Specifically, the ``User-Agent`` that the framework
          wishes us to scan.

        :param detected: Container of other information detected so-far in the
          ``User-Agent`` detection process.

        :returns: Spawned ``AgentOS`` info container describing any operating
          system information in the ``User-Agent`` in question."""

    return cls(*{
        'bsd': ('BSD', Vendor.OPEN, AgentVersion(0)),
        'linux': ('Linux', Vendor.OPEN, AgentVersion(0)),
        'macos': ('Mac OS X', Vendor.APPLE, AgentVersion(0)),
        'windows': ('Windows', Vendor.MICROSOFT, AgentVersion(0)),
        'ipad': ('iOS', Vendor.APPLE, AgentVersion(0)),
        'iphone': ('iOS', Vendor.APPLE, AgentVersion(0))
    }.get(user_agent.platform.lower().strip(), (
      'unknown', Vendor.OTHER, AgentVersion(0))))


class AgentCapabilities(AgentInfo):

  """ Holds parsed or detected information about a software HTTP client or
      browser's extra/interesting capabilities. """

  __slots__ = (
    'spdy',  # support for SPDY
    'quic',  # support for QUIC
    'webp',  # support for WebP
    'webm',  # support for WebM
    'http2'  # support for HTTP2
  )

  def __init__(self, **kwargs):

    """ Initialize this capabilities container.

        :param **kwargs: Accepts keywords for supported flags, to set them as
          active (``True``) or inactive (``False``). Currently, the supported
          capabilities flags are all ``bool`` and consist of:

          - ``spdy`` - is the client browser communicating over SPDY?
          - ``quic`` - is the client browser communicating over QUIC?
          - ``http2`` - is the client browser communicating over HTTP2?
          - ``webp`` - does the client indicate support for WebP?
          - ``webm`` - does the client indicate support for WebM?  """

    for datapoint in self.__slots__:
      setattr(self, datapoint, kwargs[datapoint] if datapoint in (
        kwargs) else None)

  @classmethod
  def scan(cls, request, user_agent, detected):

    """ Scan a target ``user_agent`` string, encapsulated by an HTTP
        ``request``, for information about an HTTP client or browser's
        indicated or implied capabilities.

        :param request: HTTP request that contains the original ``User-Agent``
          header to be scanned.

        :param user_agent: Specifically, the ``User-Agent`` that the framework
          wishes us to scan, should it be different from the original.

        :param detected: Container of other information detected so-far in the
          ``User-Agent`` scanning process.

        :returns: Spawned ``AgentCapabilities`` object describing any detected
          capabilities implied or indicated by the subject ``User-Agent``
          string. """

    detected = {}  # detected capabilities
    accept_string = request.headers['Accept'] if 'Accept' in (
      request.headers) else ''

    for datapoint, conditional in ((

      ('quic', user_agent.browser == 'chrome'),
      ('spdy', user_agent.browser in ('chrome', 'firefox', 'opera')),
      ('webm', user_agent.browser in ('chrome', 'firefox', 'opera')),
      ('webp', user_agent.browser == 'chrome' or 'webp' in accept_string))):

      detected[datapoint] = conditional
    return cls(**detected)


class AgentFingerprint(AgentInfo):

  """ Holds a full picture of detected information about a software HTTP client
      or browser, scanned or inferred from various request headers such as
      ``User-Agent`` and ``Accept``.

      Encapsulates local information about:

      - asset quality preferences indicated by browser
      - general flags for whether a client is *modern* or *ancient*
      - supported languages, character sets, mimetypes and encodings
      - a client's OS (contained in an ``AgentOS`` instance)
      - a client's inferred or indicated capabilities (in an
        ``AgentCapabilities`` instance) """

  __slots__ = (

    # == Basic Data == #
    'accept',  # full Accept request header
    'string',  # full User-Agent request header
    'vendor',  # detected Vendor of this browser (enumerated in AgentVendor)
    'version',  # detected version of this browser
    'quality',  # global relative quality preference

    # == Accept Header == #
    'charsets',  # accepted charsets
    'encodings',  # accepted encodings
    'languages',  # accepted languages
    'mimetypes',  # accepted mimetypes

    # == General Flags == #
    'modern',  # is this browser generally considered `modern`?
    'ancient',  # or is this browser considered `ancient`?

    # == Specific Browsers == #
    'chrome',  # is this Chrome or Chromium?
    'seamonkey',  # is this Seamonkey?
    'msie',  # is this Internet Exporer?
    'safari',  # is this Safari?
    'firefox',  # is this Firefox?
    'opera',  # is this Opera?
    'googlebot',  # is this google's crawler?
    'aol',  # is this AOL's crawler?
    'ask',  # is this Ask's crawler?
    'yahoo',  # is this Yahoo's crawler?

    # == Environment == #
    'mobile',  # can we detect that this is a mobile device?
    'tablet',  # can we detect that this is a tablet?
    'desktop',  # or do we fallback to desktop?
    'crawler',  # is this a known crawler?

    # == Engines == #
    'gecko',  # old mozilla/netscape engine
    'blink',  # is this the `blink` fork of webkit (Chrome)?
    'webkit',  # is this webkit (Chrome/Safari)?
    'presto',  # is this an early Opera engine, pre-Blink?
    'trident',  # always active during MSIE requests
    'spidermonkey',  # is this spidermonkey (mozilla)?

    # == Internals == #
    '__os__',  # holds an `AgentOS` object
    '__supports__')  # holds an `AgentCapabilities` object

  def __init__(self, **kwargs):

    """ Initialize a new ``AgentFingerprint`` object.

        :param **kwargs: Arbitrary container of parameters to write into the
          new ``AgentFingerprint`` object. Valid options are specified in the
          object's ``__slots__`` attribute. """

    self.__os__, self.__supports__ = None, None

    for datapoint in self.__slots__:
      setattr(self, datapoint, kwargs[datapoint] if datapoint in (
        kwargs) else None)

  @property
  def os(self):

    """ Property accessor for detected operating system information.

        :returns: ``AgentOS`` instance describing operating system information
          for a given ``AgentFingerprint`` subject. """

    return self.__os__

  @property
  def supports(self):

    """ Property accessor for inferred or indicated client capabilities.

        :returns: ``AgentCapabilities`` instance describing detected/supported
          capabilities and features for a given ``AgentFingerprint``
          subject. """

    return self.__supports__

  capabilities = supports

  @classmethod
  def scan(cls, request, ua):

    """ Scan a target HTTP ``request`` and ``User-Agent`` string for various
        pieces of information, such as an OS, browser/vendor, etc. Also scan
        other request-based headers that can provide hints about supported
        browser features and options.

        :param request: Original HTTP request providing the ``User-Agent`` to
          be scanned.

        :param ua: Specific ``User-Agent`` string requested for parsing by the
          framework.

        :returns: Spawned ``AgentFingerprint`` instance describing any and all
          information available to be parsed from the ``User-Agent`` and
          ``Accept``-series request headers. """

    detected = {
      'accept': request.headers.get('accept'),
      'string': request.headers.get('user-agent'),
      'charsets': request.accept_charsets,
      'encodings': request.accept_encodings,
      'languages': request.accept_languages,
      'mimetypes': request.accept_mimetypes
    }  # holds detected truths/guesses

    if ua is None: return cls(**{})

    # detect version first
    version = detected['version'] = ua.version.split('.')
    version_spec = []

    # take each version section as major, minor, micro
    for grouping in version:
      if len(version_spec) >= 3:
        break
      try:
        version_spec.append(int(grouping))
      except ValueError:
        break

    # if we detected *anything* as an int, add it as our version
    version = detected['version'] = AgentVersion(*tuple(
      version_spec)) if version_spec else AgentVersion(0)
    platform = ua.platform.lower().strip()

    # all others
    for datapoint, condition in ((

      ## Browser
      ('chrome', ua.browser == 'chrome'),
      ('firefox', ua.browser == 'firefox'),
      ('seamonkey', ua.browser == 'seamonkey'),
      ('safari', ua.browser == 'safari'),
      ('opera', ua.browser == 'opera'),
      ('msie', ua.browser == 'msie'),
      ('googlebot', ua.browser == 'google'),
      ('yahoo', ua.browser == 'yahoo'),
      ('aol', ua.browser == 'aol'),
      ('ask', ua.browser == 'ask'),

      ## Engines
      ('trident', ua.browser == 'msie'),
      ('blink', ua.browser in ('chrome', 'opera')),
      ('presto', ua.browser == 'opera' and version.major < 15),
      ('webkit', ua.browser in ('safari', 'chrome', 'opera')),
      ('spidermonkey', ua.browser in ('firefox', 'seamonkey')),
      ('gecko', 'Gecko' in ua.string and ('WebKit' not in ua.string and (
                                          'Chrome' not in ua.string))),

      ## Environments
      ('tablet', 'Tabl' in ua.string or 'iPad' in ua.string),
      ('crawler', ua.browser in ('google', 'yahoo', 'aol', 'ask')),
      ('mobile', 'Mobi' in ua.string or 'IEMobile' in ua.string or (
                platform in ('ios', 'iphone', 'ipad'))))):

      detected[datapoint] = condition

    # detect vendor
    detected['vendor'] = Vendor.OTHER
    for k, v in {
        Vendor.GOOGLE: detected.get('chrome') or detected.get('googlebot'),
        Vendor.MOZILLA: detected.get('firefox') or detected.get('seamonkey'),
        Vendor.MICROSOFT: detected.get('msie'),
        Vendor.OPERA: detected.get('opera'),
        Vendor.APPLE: detected.get('safari'),
        Vendor.OPEN: detected.get('seamonkey')}.iteritems():

      if v: detected['vendor'] = k

    # desktop mode
    detected['desktop'] = not any((
      detected.get('mobile'), detected.get('tablet')))

    # OS detection
    detected['__os__'] = AgentOS.scan(request, ua, detected)

    # capabilities detection
    detected['__supports__'] = AgentCapabilities.scan(request, ua, detected)

    # judge modern/ancient
    detected['modern'] = (detected['chrome'] or detected['safari'] or (
      detected['firefox'] or detected['opera'])) and (
      detected['__os__'].name in ('Mac OS X', 'Windows', 'Linux'))

    # calculate quality preferences
    detected['quality'] = {}
    if 'mimetypes' in detected:
      for mime, quality in detected['mimetypes']:
        detected['quality'][mime] = quality

    detected['ancient'] = not detected['modern']
    return cls(**detected)


@decorators.bind('http.agent')
class UserAgent(logic.Logic):

  """ Provides structured access to HTTP request headers. Interrogates values
      such as ``User-Agent`` and ``Accept`` to infer or detect things such as a
      client's OS, browser, and feature capabilities. """

  @staticmethod
  def scan(request):

    """ Scan an HTTP ``request`` for information about the other end of the
        connection. Detect as much information as possible from headers such as
        ``User-Agent`` and ``Accept``.

        :param request: HTTP request to be scanned.

        :returns: :py:class:`AgentFingerprint` instance containing any detected
          information found in the ``User-Agent`` or ``Accept``-series request
          headers. """

    return AgentFingerprint.scan(request, request.user_agent)



# -*- coding: utf-8 -*-

"""

  abstract adapter tests
  ~~~~~~~~~~~~~~~~~~~~~~

  tests abstract adapter classes, that enforce/expose
  interfaces known to the model engine proper.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import datetime

# canteen test
from canteen.test import FrameworkTest

# canteen model API
from canteen import model
from canteen.model.adapter import abstract


class TestGraphPerson(model.Vertex):

  """ simple test person object """

  name = str, {'indexed': True}


class TestGraphFriends(TestGraphPerson > TestGraphPerson):

  """ simple test friends edge """

  year_met = int, {'indexed': True}


class TestGraphGift(TestGraphPerson >> TestGraphPerson):

  """ simple directed gift edge """

  price = float, {'indexed': True}


class AbstractModelAdapterTests(FrameworkTest):

  """ Tests `model.adapter.abstract.ModelAdapter` """

  __abstract__ = True
  subject = abstract.ModelAdapter

  def _construct(self):

    """ Construct a copy of the local adapter. """

    # set to testing mode
    if hasattr(self.subject, '__testing__'):
      self.subject.__testing__ = True
    return self.subject()

  def test_abstract(self):

    """ Test `ModelAdapter` interface abstractness """

    if getattr(self, '__abstract__', False):
      with self.assertRaises(TypeError):
        self._construct()
    else:  # pragma: no cover
      self._construct()
    return getattr(self, '__abstract__', False)

  def test_utilities(self):

    """ Test `ModelAdapter` internal utilities """

    assert hasattr(self.subject, 'acquire')
    assert hasattr(self.subject, 'config')
    assert hasattr(self.subject, 'logging')
    assert hasattr(self.subject, 'serializer')
    assert hasattr(self.subject, 'encoder')
    assert hasattr(self.subject, 'compressor')

  def test_base_interface_compliance(self):

    """ Test base `ModelAdapter` interface compliance """

    assert hasattr(self.subject, 'get')
    assert hasattr(self.subject, 'put')
    assert hasattr(self.subject, 'delete')
    assert hasattr(self.subject, 'allocate_ids')
    assert hasattr(self.subject, 'encode_key')


class IndexedModelAdapterTests(AbstractModelAdapterTests):

  """ Tests `model.adapter.abstract.IndexedModelAdapter` """

  __abstract__ = True
  subject = abstract.IndexedModelAdapter

  def test_indexed_interface_compliance(self):

    """ Test `IndexedModelAdapter` interface compliance """

    assert hasattr(self.subject, 'write_indexes')
    assert hasattr(self.subject, 'clean_indexes')
    assert hasattr(self.subject, 'execute_query')

  def test_attached_indexer_compliance(self):

    """ Test `IndexedModelAdapter.Indexer` for basic functionality """

    assert hasattr(self.subject, 'Indexer')
    indexer = self.subject.Indexer

    # interrogate indexer
    assert hasattr(indexer, 'convert_key')
    assert hasattr(indexer, 'convert_date')
    assert hasattr(indexer, 'convert_time')
    assert hasattr(indexer, 'convert_datetime')

  def test_indexer_convert_key(self):

    """ Test `Indexer.convert_key` """

    indexer = self.subject.Indexer
    sample_key = model.Key('Sample', 'key')
    converted = indexer.convert_key(sample_key)

    # interrogate converted key
    assert isinstance(converted, tuple)

  def test_indexer_convert_date(self):

    """ Test `Indexer.convert_date` """

    indexer = self.subject.Indexer
    sample_date = datetime.date(year=2014, month=7, day=29)
    converted = indexer.convert_date(sample_date)

    # interrogate converted date
    assert isinstance(converted, tuple)

  def test_indexer_convert_time(self):

    """ Test `Indexer.convert_time` """

    indexer = self.subject.Indexer
    sample_time = datetime.time(hour=12, minute=30)
    converted = indexer.convert_time(sample_time)

    # interrogate converted date
    assert isinstance(converted, tuple)

  def test_indexer_convert_datetime(self):

    """ Test `Indexer.convert_datetime` """

    indexer = self.subject.Indexer
    sample_datetime = datetime.datetime(year=2014,
                                        month=7,
                                        day=29,
                                        hour=12,
                                        minute=30)
    converted = indexer.convert_datetime(sample_datetime)

    # interrogate converted date
    assert isinstance(converted, tuple)


class GraphModelAdapterTests(IndexedModelAdapterTests):

  """ Tests `model.adapter.abstract.GraphModelAdapter` """

  __abstract__ = True
  subject = abstract.GraphModelAdapter

  def test_make_vertex_nokeyname(self):

    """ Test `GraphModelAdapter` `Vertex` put with no keyname """

    if not self.test_abstract():
      t = TestGraphPerson(name="Steve")
      k = t.put(adapter=self.subject())

      assert isinstance(t, TestGraphPerson)
      assert isinstance(k, model.Key)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, (int, long))

  def test_make_vertex_keyname(self):

    """ Test `GraphModelAdapter` `Vertex` put with a keyname """

    if not self.test_abstract():
      t = TestGraphPerson(key=model.VertexKey(TestGraphPerson, "steve"),
                          name="Steve")
      k = t.put(adapter=self.subject())

      assert isinstance(t, TestGraphPerson)
      assert isinstance(k, model.Key)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, basestring)
      return t

  def test_get_vertex(self):

    """ Test `GraphModelAdapter` `Vertex` get """

    if not self.test_abstract():
      x = self.test_make_vertex_keyname()
      pulled = TestGraphPerson.get(x.key, adapter=self.subject())

      assert pulled.key == x.key
      assert isinstance(pulled.key, model.Key)
      assert isinstance(pulled.key, model.VertexKey)
      assert isinstance(pulled.key.id, basestring)

  def test_make_edge_nokeyname(self):

    """ Test `GraphModelAdapter` `Edge` put with no keyname """

    if not self.test_abstract():
      bob = TestGraphPerson(key=model.VertexKey(TestGraphPerson, "bob"),
                          name="Bob")
      k = bob.put(adapter=self.subject())

      assert isinstance(bob, TestGraphPerson)
      assert isinstance(k, model.Key)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, basestring)

      steve = self.test_make_vertex_keyname()
      f = TestGraphFriends(bob, steve)
      ek = f.put(adapter=self.subject())

      assert isinstance(ek, model.EdgeKey)
      assert isinstance(ek.id, (int, long))
      assert isinstance(f, TestGraphFriends)

  def test_make_edge_keyname(self):

    """ Test `GraphModelAdapter` `Edge` put with a keyname """

    if not self.test_abstract():

      bob = TestGraphPerson(key=model.VertexKey(TestGraphPerson, "bob"),
                            name="Bob")
      k = bob.put(adapter=self.subject())

      assert isinstance(bob, TestGraphPerson)
      assert isinstance(k, model.Key)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, basestring)

      steve = self.test_make_vertex_keyname()
      _orig_ek = model.EdgeKey(TestGraphFriends, "some-friendship")
      f = TestGraphFriends(bob, steve, key=_orig_ek)
      ek = f.put(adapter=self.subject())

      assert isinstance(ek, model.EdgeKey)
      assert isinstance(ek.id, basestring)
      assert isinstance(f, TestGraphFriends)
      assert ek.id == "some-friendship"

      return bob, steve, f

  def test_get_edge(self):

    """ Test `GraphModelAdapter` `Edge` get """

    if not self.test_abstract():
      bob, steve, friendship = self.test_make_edge_keyname()

      # fetch by key
      _f = TestGraphFriends.get(friendship.key, adapter=self.subject())

      assert isinstance(_f.key, model.Key)
      assert isinstance(_f.key, model.EdgeKey)
      assert isinstance(_f.key.id, basestring)
      assert _f.key.id == "some-friendship"
      assert _f.key.kind == "TestGraphFriends"

  def test_vertex_edges(self):

    """ Test retrieving `Edges` for a `Vertex` with `GraphModelAdapter` """

    if not self.test_abstract():
      bob, steve, friendship = self.test_make_edge_keyname()

      # friendship edge should appear for both vertexes
      _q = bob.edges(keys_only=True).fetch(adapter=self.subject(), limit=10)
      assert friendship.key in _q, (
        "expected friendship key but got:"
        " '%s' with adapter '%s'" % (
          [i for i in _q], repr(self.subject())))

      assert friendship.key in (
        steve.edges(keys_only=True).fetch(adapter=self.subject(), limit=10))
      assert "Edges" in repr(steve.edges(keys_only=True))
      assert "CONTAINS" in repr(steve.edges(keys_only=True))

  def test_vertex_neighbors(self):

    """ Test retrieving `Vertex`es for a `Vertex` with `GraphModelAdapter` """

    if not self.test_abstract():
      bob, steve, friendship = self.test_make_edge_keyname()

      # see if we can get bob's friends, which should include steve
      _q = bob.neighbors(keys_only=True).fetch(adapter=self.subject(), limit=10)
      assert steve.key in _q, (
        "failed to find steve's key in bob's neighbors."
        " instead, got '%s' for adapter '%s'" % (
          [i for i in _q], repr(self.subject())))

      # see if we can get steve's friends, which should include bob
      assert bob.key in (
        steve.neighbors(keys_only=True).fetch(adapter=self.subject(), limit=10))
      assert "Neighbors" in repr(steve.neighbors(keys_only=True))
      assert "CONTAINS" in repr(steve.neighbors(keys_only=True))


class DirectedGraphAdapterTests(GraphModelAdapterTests):

  """ Tests `model.adapter.abstract.DirectedGraphAdapter` """

  __abstract__ = True
  subject = abstract.DirectedGraphAdapter

  def test_make_directed_edge_nokeyname(self):

    """ Test saving a directed `Edge` with no keyname """

    if not self.test_abstract():
      bob = TestGraphPerson(key=model.VertexKey(TestGraphPerson, "bob"),
                          name="Bob")
      k = bob.put(adapter=self.subject())

      assert isinstance(bob, TestGraphPerson)
      assert isinstance(k, model.Key)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, basestring)

      steve = self.test_make_vertex_keyname()
      f = TestGraphGift(bob, steve)
      ek = f.put(adapter=self.subject())

      assert bob.key == f.source
      assert steve.key in f.target

      assert isinstance(ek, model.EdgeKey)
      assert isinstance(ek.id, (int, long))
      assert isinstance(f, TestGraphGift)

  def test_make_directed_edge_keyname(self):

    """ Test saving a directed `Edge` with a keyname """

    if not self.test_abstract():

      bob = TestGraphPerson(key=model.VertexKey(TestGraphPerson, "bob"),
                            name="Bob")
      k = bob.put(adapter=self.subject())

      assert isinstance(k, model.Key), (
        "instead of a key, got back the object: '%s'" % k)
      assert isinstance(bob, TestGraphPerson)
      assert isinstance(k, model.VertexKey)
      assert isinstance(k.id, basestring)

      steve = self.test_make_vertex_keyname()
      _orig_ek = model.EdgeKey(TestGraphGift, "some-gift")
      f = TestGraphGift(bob, steve, key=_orig_ek)

      ek = f.put(adapter=self.subject())

      assert isinstance(ek, model.EdgeKey)
      assert isinstance(ek.id, basestring)
      assert isinstance(f, TestGraphGift)
      assert ek.id == "some-gift"

      assert bob.key == f.source
      assert steve.key in f.target

      return bob, steve, f

  def test_get_directed_edge_nokeyname(self):

    """ Test retrieving a directed `Edge` by keyname """

    if not self.test_abstract():
      bob, steve, gift = self.test_make_directed_edge_keyname()

      # fetch by key
      _f = TestGraphGift.get(gift.key, adapter=self.subject())

      assert isinstance(_f.key, model.Key)
      assert isinstance(_f.key, model.EdgeKey)
      assert isinstance(_f.key.id, basestring)
      assert _f.key.id == "some-gift"
      assert _f.key.kind == "TestGraphGift"

  def test_edge_heads(self):

    """ Test retrieving incoming `Edge`s ending at a particular `Vertex` """

    if not self.test_abstract():
      bob, steve, gift = self.test_make_directed_edge_keyname()

      # friendship edge should appear for both vertexes
      _q = bob.edges(tails=False, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)
      assert gift.key not in _q, (
            "found gift's key among bob's edges heads, but shouldn't have."
            " instead, got: '%s' with adapter '%s'" % (
              [i for i in _q], repr(self.subject())))

      _q = steve.edges(tails=False, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)
      assert gift.key in _q, (
            "couldn't find gift's key among steve's edges heads."
            " instead, got: '%s' with adapter '%s'" % (
              [i for i in _q], repr(self.subject())))

  def test_edge_tails(self):

    """ Test retrieving outbound `Edge`s coming from a particular `Vertex` """

    if not self.test_abstract():
      bob, steve, gift = self.test_make_directed_edge_keyname()

      # friendship edge should appear for both vertexes
      _q = bob.edges(tails=True, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)
      assert gift.key in _q
      _q = steve.edges(tails=True, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)
      assert gift.key not in _q

  def test_neighbor_heads(self):

    """ Test retrieving incoming `Vertex`s ending at a particular `Vertex` """

    if not self.test_abstract():
      bob, steve, gift = self.test_make_directed_edge_keyname()

      # see if we can get steve's friends, which should include bob
      _q = steve.neighbors(tails=False, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)
      assert bob.key in _q, (
            "didn't find bob's key among steve's friends."
            " instead, got: '%s' with adapter '%s'" % (
              [i for i in _q], repr(self.subject())))

  def test_neighbor_tails(self):

    """ Test retrieving outbound `Vertex`s coming from a particular `Vertex` """

    if not self.test_abstract():
      bob, steve, gift = self.test_make_directed_edge_keyname()

      # see if we can get bob's friends, which should include steve
      assert steve.key in bob.neighbors(tails=True, keys_only=True)\
          .fetch(adapter=self.subject(), limit=10)



# -*- coding: utf-8 -*-

"""

  base RPC tests
  ~~~~~~~~~~~~~~

  tests canteen's builtin RPC systems.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""

# stdlib
import json
import inspect

# canteen
from canteen import (rpc, model)
from canteen.core import Library

# canteen testing
from canteen.test import FrameworkTest


with Library('protorpc', strict=True) as (library, protorpc):

  # load messages library
  premote, messages = (library.load('remote'),
                       library.load('messages'))


  class SampleMessage(messages.Message):

    """ Sample ProtoRPC message. """

    string = messages.StringField(1)
    integer = messages.IntegerField(2)


  class SampleModel(model.Model):

    """ Sample Canteen model. """

    string = basestring
    integer = int


  class BaseRPCTests(FrameworkTest):

    """ Tests basic `canteen.rpc` functionality """

    def test_exports(self):

      """ Test basic RPC exports """

      assert hasattr(rpc, 'VariantField')
      assert hasattr(rpc, 'StringOrIntegerField')
      assert hasattr(rpc, 'Key')
      assert hasattr(rpc, 'Echo')
      assert hasattr(rpc, 'messages')

    def test_export_classes(self):

      """ Test basic RPC class exports """

      assert hasattr(rpc, 'ServiceHandler')
      assert hasattr(rpc, 'Exception')
      assert hasattr(rpc, 'ServerException')
      assert hasattr(rpc, 'ClientException')
      assert hasattr(rpc, 'Exceptions')
      assert hasattr(rpc, 'AbstractService')
      assert hasattr(rpc, 'Service')
      assert hasattr(rpc, 'remote')

    def test_service_construction(self):

      """ Test basic `rpc.Service` construction """


      class SampleService(rpc.Service):
          """ Sample RPC service """

      SampleService()

    def test_service_mappings(self):

      """ Test generation of `rpc.Service` mappings """


      class SampleService(rpc.Service):
          """ Sample RPC service """


      class SampleServiceTwo(rpc.Service):
          """ Sample RPC service 2 """

      mappings = (
        ('/_rpc/sample1', SampleService),
        ('/_rpc/sample2', SampleServiceTwo))

      rpc.service_mappings(mappings)

    def test_service_mappings_with_dict(self):

      """ Test generation of `rpc.Service` mappings from a dict """


      class SampleService(rpc.Service):
          """ Sample RPC service """


      class SampleServiceTwo(rpc.Service):
          """ Sample RPC service 2 """

      mappings = {
        '/_rpc/sample1': SampleService,
        '/_rpc/sample2': SampleServiceTwo}

      rpc.service_mappings(mappings)

    def test_service_mappings_duplicate_uris(self):

      """ Test invalid URIs with service mappings generator """


      class SampleService(rpc.Service):
          """ Sample RPC service """


      class SampleServiceTwo(rpc.Service):
          """ Sample RPC service 2 """

      mappings = (
        ('/_rpc/sample1', SampleService),
        ('/_rpc/sample1', SampleServiceTwo))

      with self.assertRaises(premote.ServiceConfigurationError):
        rpc.service_mappings(mappings)


  class ServiceHandlerTests(FrameworkTest):

    """ Tests the `rpc.ServiceHandler` class """

    def test_construct(self):

      """ Test constructing an `rpc.ServiceHandler` """

      rpc.ServiceHandler()  # pretty simple

    def test_add_service(self):

      """ Test `ServiceHandler.add_service` """


      class SampleService(rpc.Service): pass

      # add service to handler
      handler = rpc.ServiceHandler()
      handler.add_service('sample', SampleService, **{'sample': True})

      return handler, SampleService

    def test_get_service(self):

      """ Test `ServiceHandler.get_service` """

      handler, svc = self.test_add_service()

      # try to get sample service
      sample = handler.get_service('sample')

      assert sample is svc

    def test_services_iter(self):

      """ Test `ServiceHandler.services` as an iterator """

      handler, svc = self.test_add_service()

      # iterate and make sure sample appears
      services = []
      for name, (service, config) in handler.services:
        services.append(service)

      assert svc in services, "failed to find sample in %s" % services

    def test_describe_struct(self):

      """ Test describing `Service` definitions as a dictionary """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      manifest = handler.describe()
      assert len(manifest)

      sample_item = None
      for i in manifest:
        if i[0] == 'sample':
          sample_item = i

      assert sample_item[0] == 'sample'
      assert sample_item[-1]['sample'] is True

    def test_describe_json(self):

      """ Test describing `Service` definitions as JSON """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      manifest = json.loads(handler.describe(json=True))
      assert len(manifest)

      sample_item = None
      for i in manifest:
        if i[0] == 'sample':
          sample_item = i

      assert sample_item[0] == 'sample'
      assert sample_item[-1]['sample'] is True

    def test_describe_javascript(self):

      """ Test describing `Service` definitions via JavaScript """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      manifest = handler.describe(javascript=True)
      assert 'sample' in manifest  # service name
      assert 'true' in manifest  # in config
      assert 'apptools.rpc.service.factory(' in manifest

      manifest = manifest.replace('apptools.rpc.service.factory(', '')
      manifest = manifest.replace(');', '')
      manifest = json.loads(manifest)
      assert len(manifest)

      sample_item = None
      for i in manifest:
        if i[0] == 'sample':
          sample_item = i

      assert sample_item[0] == 'sample'
      assert sample_item[-1]['sample'] is True

    def test_describe_invalid_format(self):

      """ Test describing `Service` definitions with an invalid format """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      with self.assertRaises(TypeError):
        handler.describe(json=True, javascript=True)

    def test_describe_javascript_with_custom_callable(self):

      """ Test describing `Service` definitions via JavaScript with a custom
          callable """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      manifest = handler.describe(json=False,
                                  javascript=True,
                                  callable='testing')

      assert 'sample' in manifest  # service name
      assert 'true' in manifest  # in config
      assert 'testing(' in manifest

      manifest = json.loads(manifest.replace('testing(', '').replace(');', ''))
      assert len(manifest)

      sample_item = None
      for i in manifest:
        if i[0] == 'sample':
          sample_item = i

      assert sample_item[0] == 'sample'
      assert sample_item[-1]['sample'] is True

    def test_build_wsgi_application(self):

      """ Test assembling a WSGI application for `Service` dispatch """

      handler, svc = self.test_add_service()

      # describe as dictionary and interrogate
      manifest = handler.describe(json=False, javascript=False)
      assert len(manifest)

      sample_item = None
      for i in manifest:
        if i[0] == 'sample':
          sample_item = i

      assert sample_item[0] == 'sample'
      assert sample_item[-1]['sample'] is True

      # describe as application and interrogate
      wsgi_app = handler.application

      # basic tests
      assert callable(wsgi_app)
      assert inspect.isfunction(wsgi_app)


  ## ServiceFactoryTests
  # Tests the `rpc.ServiceFactory` class.
  class ServiceFactoryTests(FrameworkTest):

    """ Tests `rpc.ServiceFactory` """

    def test_construct(self):

      """ Test construction of a new `ServiceFactory` """


      class SomeService(rpc.Service):
          """ Some sample service. """

      factory = rpc.ServiceFactory.construct(SomeService)
      assert factory
      return factory

    def test_instance_factory(self):

      """ Test instance creation through `ServiceFactory` """

      factory = self.test_construct()
      instance = factory()

      assert factory
      assert instance


  ## DecoratorTests
  # Tests the `rpc.remote` class.
  class DecoratorTests(FrameworkTest):

    """ Tests `rpc.remote` """

    def test_construct(self):

      """ Test construction of a new `rpc.remote` wrapper """

      r = rpc.remote('registered', version=1)

      assert r.name == 'registered'
      assert r.config['version'] == 1
      return r

    def test_wrap_raw(self):

      """ Test wrapping a `Service` with `rpc.remote` directly """

      r = self.test_construct()


      class RegisteredService(rpc.Service):
          """ Service mock for testing registration """

      klass = r(RegisteredService)

      assert r.target is klass
      assert rpc.ServiceHandler.get_service('registered') is klass

      return klass, r

    def test_wrap_service(self):

      """ Test wrapping a `Service` with `rpc.remote.service` """

      klass, r = self.test_wrap_raw()

      wrap = r.service(klass)
      assert callable(wrap)

    def test_wrap_method(self):

      """ Test wrapping a method with `rpc.remote.method` """


      class SampleMessage(messages.Message):

        """ Mock message """

        string = messages.StringField(1)
        integer = messages.IntegerField(2)


      class RegisteredService(rpc.Service):

        """ Service mock for testing registration """

        @rpc.remote.method(SampleMessage)
        def registered_method(self, request):  # pragma: no cover

          """ I am a registered service message. """

          request.integer += 1
          request.string += ", world!"
          return request

      assert callable(RegisteredService.registered_method)
      return RegisteredService, SampleMessage

    def test_wrap_method_models(self):

      """ Test wrapping a method with `rpc.remote.method` that uses Canteen
          models """

      class SampleRequest(model.Model):

        """ Mock request """

        string = basestring
        integer = int

      class SampleResponse(model.Model):

        """ Mock response """

        string = basestring
        integer = int

      @rpc.remote.service('registered')
      class RegisteredService(rpc.Service):

        """ Service mock for testing registration """

        @rpc.remote.method(SampleRequest, SampleResponse)
        def registered_method(self, request):  # pragma: no cover

          """ I am a registered service message. """

          return SampleResponse(**{
            'string': request.string + ', world!',
            'integer': request.integer + 1
          })

      assert callable(RegisteredService.registered_method)
      return RegisteredService, SampleRequest, SampleResponse

    def test_wrap_method_dispatch(self):

      """ Test dispatching an RPC directly through a wrapped
          `rpc.remote.method` """

      RegisteredService, SampleMessage = self.test_wrap_method()

      # make message
      request = SampleMessage(string='Hello', integer=10)

      # dispatch
      result = RegisteredService().registered_method(request)

      assert result.string == 'Hello, world!'
      assert result.integer == 11
      assert isinstance(result, SampleMessage)

    def test_wrap_method_dispatch_models(self):

      """ Test dispatching an RPC with models through a wrapped
         `rpc.remote.method` """

      RegisteredService, SampleRequest, SampleResponse = (
          self.test_wrap_method_models())

      # make message
      request = SampleRequest(string='Hello', integer=10)
      r = request.to_message()

      # dispatch
      result = RegisteredService().registered_method(r)

      assert result.string == 'Hello, world!'
      assert result.integer == 11
      assert isinstance(result, messages.Message), (
          "expected response class %s, got %s" % (messages.Message, result))



# -*- coding: utf-8 -*-

"""

  decorator utils
  ~~~~~~~~~~~~~~~

  useful (and sometimes critical) decorators, for use inside and
  outside :py:mod:`canteen`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from __future__ import print_function


# noinspection PyPep8Naming
class classproperty(property):

  """ Custom decorator for class-level property getters.
      Usable like ``@property`` and chainable with
      ``@memoize``, as long as ``@memoize`` is used as
      the inner decorator. """

  # noinspection PyMethodOverriding
  def __get__(self, instance, owner):

    """ Return the property value at the class level.

        :param instance: Current encapsulating object
        dispatching via the descriptor protocol,
        ``None`` if we are being dispatched from the
        class level.

        :param owner: Corresponding owner type, available
        whether we're dispatching at the class or instance
        level.

        :returns: Result of a ``classmethod``-wrapped,
        ``property``-decorated method. """

    return classmethod(self.fget).__get__(None, owner)()


def singleton(target):

  """ Mark a ``target`` class as a singleton, for use with the dependency
      injection system. Classes so-marked will be factoried on first-access and
      subsequently returned for all matching dependency requests.

      :param target: Target class to treat as a singleton.

      :raises RuntimeError: If something other than a class is marked for
        singleton mode.

      :returns: Decorated ``target`` class, after it has been marked. """

  if isinstance(target, type):
    setattr(target, '__singleton__', True)  # indicate this is a singleton class
    return target
  raise RuntimeError('Only classes may be marked/decorated'
                     ' as singletons. Got: "%s".' % target)  # pragma: no cover


# noinspection PyPep8Naming
class bind(object):

  """ Encapsulated binding config for an injectable meta-implementor of
      ``meta.Proxy.Component``. Allows specification of simple string names to
      be matched during dependency injection. """

  __alias__ = None  # injection alias (i.e. `source.<alias> == <target>`)
  __target__ = None  # target for injection - i.e. what should be injected
  __config__ = None  # optional *args and **kwargs to wrap ``config`` (above)
  __namespace__ = True  # do we namespace this property under it's superbind?

  def __init__(self, alias=None, namespace=True, *args, **kwargs):

    """ Initialize this binding.

        :param alias: String alias for the target object to be bound. Defaults
          to ``None``, in which case the target function or class' ``__name__``
          will be used.

        :param namespace: ``bool`` flag to activate namespacing. Used on methods
          to explicitly bind them, but namespace them under the class binding
          they are mounted from.

        :param *args:  """

    self.__alias__, self.__config__, self.__namespace__ = (
      alias,
      (args, kwargs) if (args or kwargs) else None,
      namespace)  # wrap `decorators.config` (optional)

  def __repr__(self):  # pragma: no cover

    """ Generate a pleasant string representation for this binding.

        :returns: String representation for this binding, in the format
          ``<binding 'name'>``. """

    return "<binding '%s'>" % self.__alias__ or self.__target__.__name__

  # noinspection PyUnresolvedReferences
  def __call__(self, target):

    """ Dispatch this binding (the second half of a closured decorator flow) by
        scanning the target for subbindings (if applicable) and preparing (and
        subsequently attaching) an object to describe configuration.

        :param target: Target object (usually a ``function`` or ``class``) to
          *decorate* by scanning for sub-bindings and attaching a object
          describing any injectable resources.

        :raises TypeError: If a ``target`` is passed that is not a valid
          meta-implementor of ``Proxy.Registry`` or ``Proxy.Component``.

        :returns: Decorated ``target``, after scanning for bindings and
          attaching any appropriate configuration objects. """

    from ..core import meta  # no deps in util. ever. :)

    # default to binding name
    self.__alias__ = self.__alias__ or target.__name__

    # install aliases
    target.__binding__, target.__target__, self.__target__ = (
      self, self.__alias__, target)

    # are we decorating a class?
    if isinstance(target, type):

      if issubclass(target.__class__, meta.Proxy.Registry):

        _bindings, _aliases, _hooks = set(), {}, []

        # prepare singleton if requested
        if hasattr(target, '__singleton__') and target.__singleton__:
          target.__class__.prepare(target)

        # scan for "bound" methods (bound for DI, not for Python)
        for mapping in (target.__dict__, target.__class__.__dict__):

          for k, v in mapping.iteritems():

            if k.startswith('__'): continue

            # is this a wrapped method? unwrap it
            if isinstance(v, (staticmethod, classmethod)):
              v = v.__func__  # unwrap from wrapped class/static decorator

            # is this a bound (i.e. dependency-injected) method?
            if hasattr(v, '__binding__') and v.__binding__:
              _bindings.add(k)
              if v.__binding__.__alias__:
                _aliases[v.__binding__.__alias__] = k

            # is this a hook method? register with self
            if hasattr(v, '__hooks__') and v.__hooks__:
              v.__register__(target)

        # attach bindings to target class
        target.__aliases__, target.__bindings__ = (
          _aliases, frozenset(_bindings) if _bindings else None)

        # bind locally, and internally
        return target  # args no longer supported @TODO(sgammon): look at this

      # only registry-enabled class trees can use ``bind``
      raise TypeError('Only meta-implementors of `meta.Proxy.Registry`'
                      ' (anything meta-deriving from `Registry` or `Component`'
                      ' can be bound to injection names.')  # pragma: no cover

    # allow wrapping of hook responders
    from ..core import hooks
    if self.__config__ and self.__config__[1] and (
      isinstance(self.__config__[1]['wrap'], hooks.HookResponder)):
      self.__config__[1]['wrap'].__binding__ = self

    # are we decorating a method?
    return self.__config__[1]['wrap'](target) if (
      self.__config__ and 'wrap' in self.__config__[1]) else target



# -*- coding: utf-8 -*-

"""

  HTTP agent logic tests
  ~~~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""



# -*- coding: utf-8 -*-

"""

  JSON RPC protocol
  ~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen base & core
from canteen.core import runtime
from canteen.base import protocol


## Globals
_content_types = (
  'application/json',
  'application/x-javascript',
  'text/javascript',
  'text/x-javascript',
  'text/x-json',
  'text/json')


with runtime.Library('protorpc') as (library, protorpc):

  # submodules
  protojson = library.load('protojson')
  json = __import__('json', globals(), locals(), [], 0)


  @protocol.Protocol.register('json', _content_types)
  class JSON(protocol.Protocol, protojson.ProtoJson):

    """  """

    def encode_message(self, message):

      """  """

      return protojson.ProtoJson().encode_message(message)

    def decode_message(self, message_type, encoded_message):

      """  """

      return protojson.ProtoJson().decode_message(message_type, encoded_message)



# -*- coding: utf-8 -*-

"""

  core tests
  ~~~~~~~~~~

  tests canteen's core, which contains abstract/meta code for constructing
  and gluing together the rest of canteen.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  __all__ = ('test_runtime',
             'test_injection',
             'test_meta',
             'test_hooks')
